###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test of the ProductionRequest and Production modules."""
# pylint: disable=line-too-long,invalid-name,protected-access,missing-docstring

import unittest
from unittest.mock import Mock, MagicMock

from LHCbDIRAC.BookkeepingSystem.Client.test.mock_BookkeepingClient import (
    BookkeepingClientFake,
    stepMC,
    stepMC2,
    stepStripp,
    mergeStep,
    mergeStepBHADRON,
    step1Dict,
    step2Dict,
    step124620Dict,
    step124621Dict,
    step125080,
    stepHistoMergingDict,
    mergeStepCALIBRA,
    mergeStepPIDMDST,
)
from LHCbDIRAC.ProductionManagementSystem.Client.ProductionRequest import ProductionRequest, _splitIntoProductionSteps
from LHCbDIRAC.ProductionManagementSystem.Client.Production import Production

prodsDict = {
    1: {
        "productionType": "DataStripping",
        "stepsInProd": [123, 456],
        "bkQuery": "Full",
        "removeInputsFlag": False,
        "tracking": 0,
        "outputSE": {"T1": "SE1"},
        "priority": 5,
        "cpu": 1000000,
        "outputFileMask": "",
        "outputFileStep": "",
        "input": [],
        "target": "",
        "groupSize": 1,
        "plugin": "ByRun",
        "inputDataPolicy": "download",
        "derivedProduction": 0,
        "transformationFamily": 0,
        "previousProd": None,
        "stepsInProd-ProdName": ["123['SDST']", "456['BHADRON.DST']"],
        "events": -1,
        "multicore": "False",
        "processors": (4, 4),
        "outputMode": "Any",
        "ancestorDepth": 0,
    },
    2: {
        "productionType": "Merge",
        "stepsInProd": [456],
        "bkQuery": "fromPreviousProd",
        "removeInputsFlag": True,
        "tracking": 1,
        "outputSE": {"T1": "SE1"},
        "priority": 8,
        "cpu": 300000,
        "outputFileMask": "",
        "outputFileStep": "",
        "input": [],
        "target": "",
        "groupSize": 1,
        "plugin": "BySize",
        "inputDataPolicy": "download",
        "derivedProduction": 0,
        "transformationFamily": 0,
        "previousProd": 1,
        "stepsInProd-ProdName": ["456['CALIBRATION.DST']"],
        "events": -1,
        "multicore": "False",
        "processors": (2, 2),
        "outputMode": "Local",
        "ancestorDepth": 0,
    },
    3: {
        "productionType": "Merge",
        "stepsInProd": [456],
        "bkQuery": "fromPreviousProd",
        "removeInputsFlag": True,
        "tracking": 1,
        "outputSE": {"T1": "SE1"},
        "priority": 9,
        "cpu": 10000,
        "outputFileMask": "",
        "outputFileStep": "",
        "input": [],
        "target": "",
        "groupSize": 2,
        "plugin": "BySize",
        "inputDataPolicy": "download",
        "derivedProduction": 0,
        "transformationFamily": 0,
        "previousProd": 1,
        "stepsInProd-ProdName": ["456['PID.MDST']"],
        "events": -1,
        "multicore": "True",
        "processors": (0, 2),
        "outputMode": "Any",
        "ancestorDepth": 0,
    },
    4: {
        "productionType": "Merge",
        "stepsInProd": [789],
        "bkQuery": "fromPreviousProd",
        "removeInputsFlag": True,
        "tracking": 1,
        "outputSE": {"T1": "SE1"},
        "priority": 9,
        "cpu": 1000,
        "outputFileMask": "",
        "outputFileStep": "",
        "input": [],
        "target": "",
        "groupSize": 3,
        "plugin": "BySize",
        "inputDataPolicy": "download",
        "derivedProduction": 0,
        "transformationFamily": 0,
        "previousProd": 1,
        "stepsInProd-ProdName": ["789['DAVINCIHIST', 'BRUNELHIST']"],
        "events": -1,
        "multicore": "True",
        "processors": (0, 0),
        "outputMode": "Any",
        "ancestorDepth": 0,
    },
}


class ClientTestCase(unittest.TestCase):
    """Base class for the Client test cases."""

    def setUp(self):
        self.bkClientMock = Mock()
        self.diracProdIn = Mock()
        self.diracProdIn.launchProduction.return_value = {"OK": True, "Value": 321}

        self.bkClientFake = BookkeepingClientFake()

        self.maxDiff = None


class ProductionRequestSuccess(ClientTestCase):
    def test__mcSpecialCase(self):
        # prepare the test case
        prod = Production()
        prod.setParameter("ProcessingType", "JDL", "Test", "ProductionGroupOrType")
        prod.addApplicationStep(stepDict=stepMC, modulesList=["GaudiApplication"])
        prod.addApplicationStep(stepDict=stepMC2, inputData="previousStep", modulesList=["GaudiApplication"])
        prod.priority = "1"
        prod.addFinalizationStep()
        prod.setFileMask("", "4")
        for par in prod.LHCbJob.workflow.parameters:
            if par.getName() == "Site":
                self.assertEqual(par.value, "ANY")
            if par.getName() == "outputDataFileMask":
                self.assertEqual(par.value, "")

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.tc = MagicMock()
        pr.opsH = MagicMock()
        pr.opsH.getValue.return_value = "testValue"
        prod.LHCbJob._siteSet = {"testValue"}

        pr._modifyAndLaunchMCXML(prod, {"tracking": 0})
        for par in prod.LHCbJob.workflow.parameters:
            if par.getName() == "Site":
                self.assertEqual(par.value, "testValue")
            if par.getName() == "numberOfEvents":
                self.assertEqual(par.value, "testValue")
            if par.getName() == "listoutput":
                self.assertTrue("gausshist" in dict(par.value).values())
            if par.getName() == "outputDataStep":
                self.assertEqual(par.value, "1;4")
            if par.getName() == "outputDataFileMask":
                self.assertEqual(par.value, "GAUSSHIST;DST")

        # # re-prepare the test case
        prod = Production()
        prod.setParameter("ProcessingType", "JDL", "Test", "ProductionGroupOrType")
        prod.addApplicationStep(stepDict=step1Dict, inputData="previousStep", modulesList=["GaudiApplication"])
        prod.addFinalizationStep()
        prod.setFileMask("", ["4"])

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.tc = MagicMock()
        pr.opsH = MagicMock()
        pr.opsH.getValue.return_value = "testValue"
        prod.LHCbJob._siteSet = {"testValue"}

        # # re-run
        pr._mcSpecialCase(prod, {"tracking": 0})
        for par in prod.LHCbJob.workflow.parameters:
            if par.getName() == "Site":
                self.assertEqual(par.value, "ANY")
            if par.getName() == "listoutput":
                self.assertTrue("gausshist" not in dict(par.value).values())
            if par.getName() == "outputDataStep":
                self.assertEqual(par.value, "4")
            if par.getName() == "outputDataFileMask":
                self.assertEqual(par.value, "")
            if par.getName() == "Priority":
                self.assertEqual(par.value, "1")

        # And again re-prepare the test case
        prod = Production()
        prod.setParameter("ProcessingType", "JDL", "Test", "ProductionGroupOrType")
        prod.addApplicationStep(stepDict=step1Dict, inputData="previousStep", modulesList=["GaudiApplication"])
        prod.addFinalizationStep()
        prod.setFileMask("")

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.tc = MagicMock()
        pr.opsH = MagicMock()
        pr.opsH.getValue.return_value = "testValue"
        prod.LHCbJob._siteSet = {"testValue"}

        pr._mcSpecialCase(prod, {"tracking": 0})
        for par in prod.LHCbJob.workflow.parameters:
            if par.getName() == "Site":
                self.assertEqual(par.value, "ANY")
            if par.getName() == "listoutput":
                self.assertTrue("gausshist" not in dict(par.value).values())
            if par.getName() == "outputDataFileMask":
                self.assertEqual(par.value, "")
            if par.getName() == "Priority":
                self.assertEqual(par.value, "1")

    def test_resolveStepsSuccess(self):
        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.stepsList = ["123"]
        pr.outputVisFlag = [{"1": "N"}]
        pr.compressionLvl = ["LOW"]
        pr.resolveSteps()
        expected = dict(step1Dict)
        expected["OptionFiles"] = "optsFiles;$APPCONFIGOPTS/Persistency/Compression-ZLIB-1.py"
        self.assertEqual(pr.stepsListDict, [expected])

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.stepsList = ["125080"]  # this one already has a compression level
        pr.outputVisFlag = [{"1": "N"}]
        pr.resolveSteps()
        expected = dict(step125080)
        expectedOptionFiles = "$APPCONFIGOPTS/Gauss/Sim08-Beam4000GeV-mu100-2012-nu2.5.py;"
        expectedOptionFiles += "$DECFILESROOT/options/11102400.py;$LBPYTHIA8ROOT/options/Pythia8.py;"
        expectedOptionFiles += "$APPCONFIGOPTS/Gauss/G4PL_FTFP_BERT_EmNoCuts.py;"
        expected["OptionFiles"] = expectedOptionFiles
        self.assertEqual(pr.stepsListDict, [expected])

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.stepsList = ["125080"]  # this one already has a compression level
        pr.compressionLvl = ["LOW"]
        pr.outputVisFlag = [{"1": "N"}]
        pr.resolveSteps()
        expected = dict(step125080)
        self.assertEqual(pr.stepsListDict, [expected])

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.stepsList = ["125080"]  # this one already has a compression level
        pr.compressionLvl = ["HIGH"]  # this will replace it
        pr.outputVisFlag = [{"1": "N"}]
        pr.resolveSteps()
        expected = dict(step125080)
        expectedOptionFiles = "$APPCONFIGOPTS/Gauss/Sim08-Beam4000GeV-mu100-2012-nu2.5.py;"
        expectedOptionFiles += "$DECFILESROOT/options/11102400.py;"
        expectedOptionFiles += "$LBPYTHIA8ROOT/options/Pythia8.py;"
        expectedOptionFiles += "$APPCONFIGOPTS/Gauss/G4PL_FTFP_BERT_EmNoCuts.py;"
        expected["OptionFiles"] = expectedOptionFiles
        self.assertEqual(pr.stepsListDict, [expected])

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.stepsList = ["123", "456"]
        pr.compressionLvl = ["LOW", "HIGH"]
        pr.outputVisFlag = [{"1": "N"}, {"2": "N"}]
        expected1 = dict(step1Dict)
        expected1["OptionFiles"] = "optsFiles;$APPCONFIGOPTS/Persistency/Compression-ZLIB-1.py"
        expected2 = dict(step2Dict)
        expected2["OptionFiles"] = "optsFiles;$APPCONFIGOPTS/Persistency/Compression-LZMA-4.py"
        pr.resolveSteps()
        self.assertEqual(pr.stepsListDict, [expected1, expected2])

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.stepsList = ["123", "456", "789"]
        pr.compressionLvl = ["LOW", "HIGH", "HIGH"]
        pr.outputVisFlag = [{"1": "N"}, {"2": "N"}, {"3": "Y"}]
        expected1 = dict(step1Dict)
        expected1["OptionFiles"] = "optsFiles;$APPCONFIGOPTS/Persistency/Compression-ZLIB-1.py"
        expected2 = dict(step2Dict)
        expected2["OptionFiles"] = "optsFiles;$APPCONFIGOPTS/Persistency/Compression-LZMA-4.py"
        expected3 = dict(stepHistoMergingDict)
        expected3["OptionFiles"] = "optsFiles;$APPCONFIGOPTS/Persistency/Compression-LZMA-4.py"
        pr.resolveSteps()
        self.assertEqual(pr.stepsListDict, [expected1, expected2, expected3])

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.stepsList = ["124620", "124621"]
        pr.compressionLvl = ["LOW", "HIGH"]
        pr.outputVisFlag = [{"1": "N"}, {"2": "Y"}]
        expected1 = dict(step124620Dict)
        expected1["OptionFiles"] = "$APPCONFIGOPTS/Boole/Default.py;$APPCONFIGOPTS/Boole/DataType-2012.py;"
        expected2 = dict(step124621Dict)
        expected2["OptionFiles"] = "someBrunelOptions"
        pr.resolveSteps()
        self.assertEqual(pr.stepsListDict, [expected1, expected2])

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.stepsList = ["123", "456", "", ""]
        pr.outputVisFlag = [{"1": "N"}, {"2": "N"}]
        pr.compressionLvl = ["LOW", "LOW", "HIGH", "HIGH"]
        pr.resolveSteps()
        self.assertEqual(
            pr.stepsListDict,
            [
                {
                    "StepId": 123,
                    "StepName": "Stripping14-Stripping",
                    "ApplicationName": "DaVinci",
                    "ApplicationVersion": "v2r2",
                    "ExtraOptions": "",
                    "EventTimeout": "",
                    "OptionFiles": "optsFiles;$APPCONFIGOPTS/Persistency/Compression-ZLIB-1.py",
                    "Visible": "Yes",
                    "ExtraPackages": "eps",
                    "ProcessingPass": "procPass",
                    "OptionsFormat": "",
                    "SystemConfig": "",
                    "DDDB": "",
                    "CONDDB": "123456",
                    "DQTag": "",
                    "isMulticore": "N",
                    "prodStepID": "123['SDST']",
                    "mcTCK": "",
                    "visibilityFlag": [
                        {"Visible": "N", "FileType": "BHADRON.DST"},
                        {"Visible": "N", "FileType": "CALIBRATION.DST"},
                    ],
                    "fileTypesIn": ["SDST"],
                    "fileTypesOut": ["BHADRON.DST", "CALIBRATION.DST"],
                },
                {
                    "StepId": 456,
                    "StepName": "Merge",
                    "ApplicationName": "LHCb",
                    "ApplicationVersion": "v1r2",
                    "ExtraOptions": "",
                    "EventTimeout": "",
                    "OptionFiles": "optsFiles;$APPCONFIGOPTS/Persistency/Compression-ZLIB-1.py",
                    "Visible": "Yes",
                    "ExtraPackages": "eps",
                    "ProcessingPass": "procPass",
                    "OptionsFormat": "",
                    "SystemConfig": "x86_64-slc6-gcc49-opt",
                    "prodStepID": "456['BHADRON.DST', 'CALIBRATION.DST']",
                    "DDDB": "",
                    "CONDDB": "123456",
                    "DQTag": "",
                    "isMulticore": "N",
                    "mcTCK": "",
                    "visibilityFlag": [
                        {"Visible": "N", "FileType": "BHADRON.DST"},
                        {"Visible": "N", "FileType": "CALIBRATION.DST"},
                    ],
                    "fileTypesIn": ["BHADRON.DST", "CALIBRATION.DST"],
                    "fileTypesOut": ["BHADRON.DST", "CALIBRATION.DST"],
                },
            ],
        )

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.stepsList = ["123"]
        pr.outputVisFlag = [{"1": "N"}]
        pr.resolveSteps()
        self.assertEqual(
            pr.stepsListDict,
            [
                {
                    "StepId": 123,
                    "StepName": "Stripping14-Stripping",
                    "ApplicationName": "DaVinci",
                    "ApplicationVersion": "v2r2",
                    "ExtraOptions": "",
                    "EventTimeout": "",
                    "OptionFiles": "optsFiles",
                    "Visible": "Yes",
                    "ExtraPackages": "eps",
                    "ProcessingPass": "procPass",
                    "OptionsFormat": "",
                    "SystemConfig": "",
                    "DDDB": "",
                    "CONDDB": "123456",
                    "DQTag": "",
                    "isMulticore": "N",
                    "prodStepID": "123['SDST']",
                    "mcTCK": "",
                    "visibilityFlag": [
                        {"Visible": "N", "FileType": "BHADRON.DST"},
                        {"Visible": "N", "FileType": "CALIBRATION.DST"},
                    ],
                    "fileTypesIn": ["SDST"],
                    "fileTypesOut": ["BHADRON.DST", "CALIBRATION.DST"],
                }
            ],
        )
        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.compressionLvl = ["HIGH"]
        pr.stepsList = ["123"]
        pr.outputVisFlag = [{"1": "N"}]
        pr.resolveSteps()
        self.assertEqual(
            pr.stepsListDict,
            [
                {
                    "StepId": 123,
                    "StepName": "Stripping14-Stripping",
                    "ApplicationName": "DaVinci",
                    "ApplicationVersion": "v2r2",
                    "ExtraOptions": "",
                    "EventTimeout": "",
                    "ProcessingPass": "procPass",
                    "OptionsFormat": "",
                    "SystemConfig": "",
                    "OptionFiles": "optsFiles;$APPCONFIGOPTS/Persistency/Compression-LZMA-4.py",
                    "Visible": "Yes",
                    "ExtraPackages": "eps",
                    "prodStepID": "123['SDST']",
                    "mcTCK": "",
                    "visibilityFlag": [
                        {"Visible": "N", "FileType": "BHADRON.DST"},
                        {"Visible": "N", "FileType": "CALIBRATION.DST"},
                    ],
                    "DDDB": "",
                    "CONDDB": "123456",
                    "DQTag": "",
                    "isMulticore": "N",
                    "fileTypesIn": ["SDST"],
                    "fileTypesOut": ["BHADRON.DST", "CALIBRATION.DST"],
                }
            ],
        )

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.compressionLvl = ["HIGH"]
        pr.outputVisFlag = [{"1": "Y"}]
        # pr.outputVisFlag = [{'Visibility':'N','FileType':'BHADRON.DST'},{'Visibility':'N','FileType':'CALIBRATION.DST'}]
        pr.stepsList = ["999"]
        pr.resolveSteps()
        self.assertEqual(
            pr.stepsListDict,
            [
                {
                    "StepId": 999,
                    "StepName": "Stripping28",
                    "ApplicationName": "DaVinci",
                    "ApplicationVersion": "v41r3",
                    "ExtraOptions": "",
                    "EventTimeout": "",
                    # 'OptionFiles':'$APPCONFIGOPTS/Persistency/Compression-ZLIB-1.py;$APPCONFIGOPTS/Persistency/Compression-LZMA-4.py',
                    # 'OptionFiles':'$APPCONFIGOPTS/Persistency/Compression-LZMA-4.py',
                    "OptionFiles": "",
                    "Visible": "Yes",
                    "Usable": "Yes",
                    "ProcessingPass": "Stripping28",
                    "SystemConfig": "x86_64-slc5-gcc49-opt",
                    "ExtraPackages": "AppConfig.v3r306",
                    "mcTCK": "",
                    "prodStepID": "999['SDST']",
                    "DDDB": "dddb-20150724",
                    "CONDDB": "cond-20161011",
                    "DQTag": "",
                    "isMulticore": "N",
                    "fileTypesIn": ["SDST"],
                    "visibilityFlag": [
                        {"Visible": "Y", "FileType": "BHADRON.DST"},
                        {"Visible": "Y", "FileType": "CALIBRATION.DST"},
                    ],
                    "fileTypesOut": ["BHADRON.DST", "CALIBRATION.DST"],
                }
            ],
        )

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.compressionLvl = ["HIGH"]
        pr.outputVisFlag = [{"1": "Y"}]
        pr.stepsList = ["998"]
        pr.resolveSteps()
        self.assertEqual(
            pr.stepsListDict,
            [
                {
                    "StepId": 998,
                    "StepName": "Stripping28",
                    "ApplicationName": "DaVinci",
                    "ApplicationVersion": "v41r3",
                    "ExtraOptions": "",
                    "EventTimeout": "",
                    # 'OptionFiles':'$APPCONFIGOPTS/Persistency/Compression-ZLIB-1.py;$APPCONFIGOPTS/Persistency/Compression-LZMA-4.py',
                    "OptionFiles": "$APPCONFIGOPTS/Persistency/Compression-LZMA-4.py",
                    "Visible": "Yes",
                    "Usable": "Yes",
                    "ProcessingPass": "Stripping28",
                    "SystemConfig": "x86_64-slc6-gcc49-opt",
                    "ExtraPackages": "AppConfig.v3r306",
                    "mcTCK": "",
                    "prodStepID": "998['SDST']",
                    "DDDB": "dddb-20150724",
                    "CONDDB": "cond-20161011",
                    "DQTag": "",
                    "isMulticore": "N",
                    "fileTypesIn": ["SDST"],
                    "visibilityFlag": [
                        {"Visible": "Y", "FileType": "BHADRON.DST"},
                        {"Visible": "Y", "FileType": "CALIBRATION.DST"},
                    ],
                    "fileTypesOut": ["BHADRON.DST", "CALIBRATION.DST"],
                }
            ],
        )

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.stepsList = ["999"]
        pr.outputVisFlag = [{"1": "Y"}]
        pr.compressionLvl = [""] * 4
        pr.resolveSteps()
        self.assertEqual(
            pr.stepsListDict,
            [
                {
                    "StepId": 999,
                    "StepName": "Stripping28",
                    "ApplicationName": "DaVinci",
                    "ApplicationVersion": "v41r3",
                    "ExtraOptions": "",
                    "EventTimeout": "",
                    # 'OptionFiles':'$APPCONFIGOPTS/Persistency/Compression-ZLIB-1.py;$APPCONFIGOPTS/Persistency/Compression-LZMA-4.py',
                    "OptionFiles": "",
                    "Visible": "Yes",
                    "Usable": "Yes",
                    "ProcessingPass": "Stripping28",
                    "SystemConfig": "x86_64-slc5-gcc49-opt",
                    "ExtraPackages": "AppConfig.v3r306",
                    "mcTCK": "",
                    "prodStepID": "999['SDST']",
                    "DDDB": "dddb-20150724",
                    "CONDDB": "cond-20161011",
                    "DQTag": "",
                    "isMulticore": "N",
                    "fileTypesIn": ["SDST"],
                    "visibilityFlag": [
                        {"Visible": "Y", "FileType": "BHADRON.DST"},
                        {"Visible": "Y", "FileType": "CALIBRATION.DST"},
                    ],
                    "fileTypesOut": ["BHADRON.DST", "CALIBRATION.DST"],
                }
            ],
        )

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.compressionLvl = ["HIGH"]
        pr.outputVisFlag = [{"1": "Y"}]
        pr.stepsList = ["997"]
        pr.resolveSteps()
        optionFilesStep997 = "$APPCONFIGOPTS/Brunel/DataType-2016.py;"
        optionFilesStep997 += "$APPCONFIGOPTS/Brunel/PbPb-GECs.py;$APPCONFIGOPTS/Persistency/Compression-LZMA-4.py"
        self.assertEqual(
            pr.stepsListDict,
            [
                {
                    "StepId": 997,
                    "StepName": "Stripping28",
                    "ApplicationName": "DaVinci",
                    "ApplicationVersion": "v41r3",
                    "ExtraOptions": "",
                    "EventTimeout": "",
                    "OptionFiles": optionFilesStep997,
                    "Visible": "Yes",
                    "Usable": "Yes",
                    "ProcessingPass": "Stripping28",
                    "SystemConfig": "x86_64-slc6-gcc49-opt",
                    "ExtraPackages": "AppConfig.v3r306",
                    "mcTCK": "",
                    "prodStepID": "997['SDST']",
                    "DDDB": "dddb-20150724",
                    "CONDDB": "cond-20161011",
                    "DQTag": "",
                    "isMulticore": "N",
                    "fileTypesIn": ["SDST"],
                    "visibilityFlag": [
                        {"Visible": "Y", "FileType": "BHADRON.DST"},
                        {"Visible": "Y", "FileType": "CALIBRATION.DST"},
                    ],
                    "fileTypesOut": ["BHADRON.DST", "CALIBRATION.DST"],
                }
            ],
        )

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.compressionLvl = ["HIGH"]
        pr.outputVisFlag = [{"1": "Y"}]
        pr.stepsList = ["996"]
        pr.resolveSteps()
        optionFilesStep996 = "$APPCONFIGOPTS/Brunel/DataType-2016.py;"
        optionFilesStep996 += "$APPCONFIGOPTS/Persistency/Compression-LZMA-4.py;$APPCONFIGOPTS/Brunel/PbPb-GECs.py"
        self.assertEqual(
            pr.stepsListDict,
            [
                {
                    "StepId": 996,
                    "StepName": "Stripping28",
                    "ApplicationName": "DaVinci",
                    "ApplicationVersion": "v41r3",
                    "ExtraOptions": "",
                    "EventTimeout": "",
                    "OptionFiles": optionFilesStep996,
                    "Visible": "Yes",
                    "Usable": "Yes",
                    "ProcessingPass": "Stripping28",
                    "SystemConfig": "x86_64-slc6-gcc49-opt",
                    "ExtraPackages": "AppConfig.v3r306",
                    "mcTCK": "",
                    "prodStepID": "996['SDST']",
                    "DDDB": "dddb-20150724",
                    "CONDDB": "cond-20161011",
                    "DQTag": "",
                    "isMulticore": "N",
                    "fileTypesIn": ["SDST"],
                    "visibilityFlag": [
                        {"Visible": "Y", "FileType": "BHADRON.DST"},
                        {"Visible": "Y", "FileType": "CALIBRATION.DST"},
                    ],
                    "fileTypesOut": ["BHADRON.DST", "CALIBRATION.DST"],
                }
            ],
        )

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.compressionLvl = [""] * 4
        pr.outputVisFlag = [{"1": "Y"}]
        pr.stepsList = ["1098"]
        pr.resolveSteps()
        self.assertEqual(
            pr.stepsListDict,
            [
                {
                    "StepId": 1098,
                    "StepName": "Stripping28",
                    "ApplicationName": "DaVinci",
                    "ApplicationVersion": "v41r3",
                    "ExtraOptions": "",
                    "EventTimeout": "",
                    "OptionFiles": "$APPCONFIGOPTS/Brunel/DataType-2016.py;",
                    "Visible": "Yes",
                    "Usable": "Yes",
                    "ProcessingPass": "Stripping28",
                    "SystemConfig": "x86_64-slc6-gcc49-opt",
                    "ExtraPackages": "AppConfig.v3r306",
                    "mcTCK": "",
                    "prodStepID": "1098['SDST']",
                    "DDDB": "dddb-20150724",
                    "CONDDB": "cond-20161011",
                    "DQTag": "",
                    "isMulticore": "N",
                    "fileTypesIn": ["SDST"],
                    "visibilityFlag": [
                        {"Visible": "Y", "FileType": "BHADRON.DST"},
                        {"Visible": "Y", "FileType": "CALIBRATION.DST"},
                    ],
                    "fileTypesOut": ["BHADRON.DST", "CALIBRATION.DST"],
                }
            ],
        )

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.compressionLvl = ["HIGH"]
        pr.outputVisFlag = [{"1": "Y"}]
        pr.specialOutputVisFlag = [{"1": {"BHADRON.DST": "Y"}}]
        pr.stepsList = ["1098"]
        pr.resolveSteps()
        optionFilesStep1098 = "$APPCONFIGOPTS/Brunel/DataType-2016.py;"
        optionFilesStep1098 += "$APPCONFIGOPTS/Persistency/Compression-LZMA-4.py;"
        optionFilesStep1098 += "$APPCONFIGOPTS/Persistency/Compression-LZMA-4.py"
        self.assertEqual(
            pr.stepsListDict,
            [
                {
                    "StepId": 1098,
                    "StepName": "Stripping28",
                    "ApplicationName": "DaVinci",
                    "ApplicationVersion": "v41r3",
                    "ExtraOptions": "",
                    "EventTimeout": "",
                    "OptionFiles": optionFilesStep1098,
                    "Visible": "Yes",
                    "Usable": "Yes",
                    "ProcessingPass": "Stripping28",
                    "SystemConfig": "x86_64-slc6-gcc49-opt",
                    "ExtraPackages": "AppConfig.v3r306",
                    "mcTCK": "",
                    "prodStepID": "1098['SDST']",
                    "DDDB": "dddb-20150724",
                    "CONDDB": "cond-20161011",
                    "DQTag": "",
                    "isMulticore": "N",
                    "fileTypesIn": ["SDST"],
                    "visibilityFlag": [
                        {"Visible": "Y", "FileType": "BHADRON.DST"},
                        {"Visible": "Y", "FileType": "CALIBRATION.DST"},
                    ],
                    "fileTypesOut": ["BHADRON.DST", "CALIBRATION.DST"],
                }
            ],
        )

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.compressionLvl = [""]
        pr.outputVisFlag = [{"1": "Y"}]
        pr.specialOutputVisFlag = [{"1": {"BHADRON.DST": "Y"}}]
        pr.stepsList = ["1099"]
        pr.resolveSteps()
        self.assertEqual(
            pr.stepsListDict,
            [
                {
                    "StepId": 1099,
                    "StepName": "Stripping28",
                    "ApplicationName": "DaVinci",
                    "ApplicationVersion": "v41r3",
                    "ExtraOptions": "",
                    "EventTimeout": "",
                    "OptionFiles": "$APPCONFIGOPTS/Brunel/DataType-2016.py;",
                    "Visible": "Yes",
                    "Usable": "Yes",
                    "ProcessingPass": "Stripping28",
                    "SystemConfig": "x86_64-slc6-gcc49-opt",
                    "ExtraPackages": "AppConfig.v3r306",
                    "mcTCK": "",
                    "prodStepID": "1099['SDST']",
                    "DDDB": "dddb-20150724",
                    "CONDDB": "cond-20161011",
                    "DQTag": "",
                    "isMulticore": "N",
                    "fileTypesIn": ["SDST"],
                    "visibilityFlag": [
                        {"Visible": "Y", "FileType": "BHADRON.DST"},
                        {"Visible": "Y", "FileType": "CALIBRATION.DST"},
                    ],
                    "fileTypesOut": ["BHADRON.DST", "CALIBRATION.DST"],
                }
            ],
        )

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.compressionLvl = ["", "HIGH", "LOW"]
        pr.outputVisFlag = [{"1": "N", "2": "N", "3": "N"}]
        pr.specialOutputVisFlag = [{"3": {"BHADRON.DST": "Y"}}]
        pr.stepsList = ["1098", "996", "997"]
        pr.resolveSteps()
        optionFilesStep996 = "$APPCONFIGOPTS/Brunel/DataType-2016.py;"
        optionFilesStep996 += "$APPCONFIGOPTS/Persistency/Compression-LZMA-4.py;"
        optionFilesStep996 += "$APPCONFIGOPTS/Brunel/PbPb-GECs.py"
        optionFilesStep997 = "$APPCONFIGOPTS/Brunel/DataType-2016.py;"
        optionFilesStep997 += "$APPCONFIGOPTS/Brunel/PbPb-GECs.py;"
        optionFilesStep997 += "$APPCONFIGOPTS/Persistency/Compression-ZLIB-1.py"
        self.assertEqual(
            pr.stepsListDict,
            [
                {
                    "StepId": 1098,
                    "StepName": "Stripping28",
                    "ApplicationName": "DaVinci",
                    "ApplicationVersion": "v41r3",
                    "ExtraOptions": "",
                    "EventTimeout": "",
                    "OptionFiles": "$APPCONFIGOPTS/Brunel/DataType-2016.py;",
                    "Visible": "Yes",
                    "Usable": "Yes",
                    "ProcessingPass": "Stripping28",
                    "SystemConfig": "x86_64-slc6-gcc49-opt",
                    "ExtraPackages": "AppConfig.v3r306",
                    "mcTCK": "",
                    "prodStepID": "1098['SDST']",
                    "DDDB": "dddb-20150724",
                    "CONDDB": "cond-20161011",
                    "DQTag": "",
                    "isMulticore": "N",
                    "fileTypesIn": ["SDST"],
                    "visibilityFlag": [
                        {"Visible": "N", "FileType": "BHADRON.DST"},
                        {"Visible": "N", "FileType": "CALIBRATION.DST"},
                    ],
                    "fileTypesOut": ["BHADRON.DST", "CALIBRATION.DST"],
                },
                {
                    "StepId": 996,
                    "StepName": "Stripping28",
                    "ApplicationName": "DaVinci",
                    "ApplicationVersion": "v41r3",
                    "ExtraOptions": "",
                    "EventTimeout": "",
                    "OptionFiles": optionFilesStep996,
                    "Visible": "Yes",
                    "Usable": "Yes",
                    "ProcessingPass": "Stripping28",
                    "SystemConfig": "x86_64-slc6-gcc49-opt",
                    "ExtraPackages": "AppConfig.v3r306",
                    "mcTCK": "",
                    "prodStepID": "996['SDST']",
                    "DDDB": "dddb-20150724",
                    "CONDDB": "cond-20161011",
                    "DQTag": "",
                    "isMulticore": "N",
                    "fileTypesIn": ["SDST"],
                    "visibilityFlag": [
                        {"Visible": "N", "FileType": "BHADRON.DST"},
                        {"Visible": "N", "FileType": "CALIBRATION.DST"},
                    ],
                    "fileTypesOut": ["BHADRON.DST", "CALIBRATION.DST"],
                },
                {
                    "StepId": 997,
                    "StepName": "Stripping28",
                    "ApplicationName": "DaVinci",
                    "ApplicationVersion": "v41r3",
                    "ExtraOptions": "",
                    "EventTimeout": "",
                    "OptionFiles": optionFilesStep997,
                    "Visible": "Yes",
                    "Usable": "Yes",
                    "ProcessingPass": "Stripping28",
                    "SystemConfig": "x86_64-slc6-gcc49-opt",
                    "ExtraPackages": "AppConfig.v3r306",
                    "mcTCK": "",
                    "prodStepID": "997['SDST']",
                    "DDDB": "dddb-20150724",
                    "CONDDB": "cond-20161011",
                    "DQTag": "",
                    "isMulticore": "N",
                    "fileTypesIn": ["SDST"],
                    "visibilityFlag": [
                        {"Visible": "Y", "FileType": "BHADRON.DST"},
                        {"Visible": "N", "FileType": "CALIBRATION.DST"},
                    ],
                    "fileTypesOut": ["BHADRON.DST", "CALIBRATION.DST"],
                },
            ],
        )

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.compressionLvl = ["", "HIGH"]
        pr.outputVisFlag = [{"1": "N", "2": "N", "3": "N"}]
        pr.specialOutputVisFlag = [{"2": {"BHADRON.DST": "Y"}}]
        pr.stepsList = ["1098", "996", "997"]
        pr.resolveSteps()
        optionFilesStep996 = "$APPCONFIGOPTS/Brunel/DataType-2016.py;"
        optionFilesStep996 += "$APPCONFIGOPTS/Persistency/Compression-LZMA-4.py;$APPCONFIGOPTS/Brunel/PbPb-GECs.py"
        optionFilesStep997 = "$APPCONFIGOPTS/Brunel/DataType-2016.py;$APPCONFIGOPTS/Brunel/PbPb-GECs.py"
        self.assertEqual(
            pr.stepsListDict,
            [
                {
                    "StepId": 1098,
                    "StepName": "Stripping28",
                    "ApplicationName": "DaVinci",
                    "ApplicationVersion": "v41r3",
                    "ExtraOptions": "",
                    "EventTimeout": "",
                    "OptionFiles": "$APPCONFIGOPTS/Brunel/DataType-2016.py;",
                    "Visible": "Yes",
                    "Usable": "Yes",
                    "ProcessingPass": "Stripping28",
                    "SystemConfig": "x86_64-slc6-gcc49-opt",
                    "ExtraPackages": "AppConfig.v3r306",
                    "mcTCK": "",
                    "prodStepID": "1098['SDST']",
                    "DDDB": "dddb-20150724",
                    "CONDDB": "cond-20161011",
                    "DQTag": "",
                    "isMulticore": "N",
                    "fileTypesIn": ["SDST"],
                    "visibilityFlag": [
                        {"Visible": "N", "FileType": "BHADRON.DST"},
                        {"Visible": "N", "FileType": "CALIBRATION.DST"},
                    ],
                    "fileTypesOut": ["BHADRON.DST", "CALIBRATION.DST"],
                },
                {
                    "StepId": 996,
                    "StepName": "Stripping28",
                    "ApplicationName": "DaVinci",
                    "ApplicationVersion": "v41r3",
                    "ExtraOptions": "",
                    "EventTimeout": "",
                    "OptionFiles": optionFilesStep996,
                    "Visible": "Yes",
                    "Usable": "Yes",
                    "ProcessingPass": "Stripping28",
                    "SystemConfig": "x86_64-slc6-gcc49-opt",
                    "ExtraPackages": "AppConfig.v3r306",
                    "mcTCK": "",
                    "prodStepID": "996['SDST']",
                    "DDDB": "dddb-20150724",
                    "CONDDB": "cond-20161011",
                    "DQTag": "",
                    "isMulticore": "N",
                    "fileTypesIn": ["SDST"],
                    "visibilityFlag": [
                        {"Visible": "Y", "FileType": "BHADRON.DST"},
                        {"Visible": "N", "FileType": "CALIBRATION.DST"},
                    ],
                    "fileTypesOut": ["BHADRON.DST", "CALIBRATION.DST"],
                },
                {
                    "StepId": 997,
                    "StepName": "Stripping28",
                    "ApplicationName": "DaVinci",
                    "ApplicationVersion": "v41r3",
                    "ExtraOptions": "",
                    "EventTimeout": "",
                    "OptionFiles": optionFilesStep997,
                    "Visible": "Yes",
                    "Usable": "Yes",
                    "ProcessingPass": "Stripping28",
                    "SystemConfig": "x86_64-slc6-gcc49-opt",
                    "ExtraPackages": "AppConfig.v3r306",
                    "mcTCK": "",
                    "prodStepID": "997['SDST']",
                    "DDDB": "dddb-20150724",
                    "CONDDB": "cond-20161011",
                    "DQTag": "",
                    "isMulticore": "N",
                    "fileTypesIn": ["SDST"],
                    "visibilityFlag": [
                        {"Visible": "N", "FileType": "BHADRON.DST"},
                        {"Visible": "N", "FileType": "CALIBRATION.DST"},
                    ],
                    "fileTypesOut": ["BHADRON.DST", "CALIBRATION.DST"],
                },
            ],
        )

    def test__determineOutputSEs(self):
        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.stepsList = ["123", "456"]
        pr.compressionLvl = ["LOW", "HIGH"]
        pr.outputVisFlag = [{"1": "N"}, {"2": "N"}]
        pr.resolveSteps()
        pr.outputSEs = ["SE1", "SE2"]
        pr.specialOutputSEs = [{}, {}]
        pr._determineOutputSEs()
        self.assertEqual(
            pr.outputSEsPerFileType,
            [{"CALIBRATION.DST": "SE1", "BHADRON.DST": "SE1"}, {"CALIBRATION.DST": "SE2", "BHADRON.DST": "SE2"}],
        )

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.outputVisFlag = [{"1": "N"}, {"2": "N"}]
        pr.stepsList = ["123", "456"]
        pr.compressionLvl = ["LOW", "HIGH"]
        pr.resolveSteps()
        pr.outputSEs = ["SE1", "SE2"]
        pr.specialOutputSEs = [{"CALIBRATION.DST": "SE3"}, {}]
        pr._determineOutputSEs()
        self.assertEqual(
            pr.outputSEsPerFileType,
            [{"CALIBRATION.DST": "SE3", "BHADRON.DST": "SE1"}, {"CALIBRATION.DST": "SE2", "BHADRON.DST": "SE2"}],
        )

    def test__applyOptionalCorrections(self):
        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.prodsTypeList = ["DataStripping", "Merge"]
        pr.plugins = ["ByRun", "BySize"]
        pr.stepsListDict = [stepStripp, mergeStep]
        pr.outputVisFlag = [{"1": "N"}, {"2": "N"}]
        pr.stepsList = [123, 456]
        pr.stepsInProds = [[1], [2]]
        pr.outputSEs = ["Tier1-DST", "Tier1-M-DST"]
        pr.specialOutputSEs = [{}, {}]
        pr.priorities = [1, 4]
        pr.cpus = [10, 100]
        pr.groupSizes = [1, 2]
        pr.bkFileType = "TYPE"
        pr.removeInputsFlags = [False, True]
        pr.outputFileMasks = ["", "dst"]
        pr.outputFileSteps = ["", "2"]
        pr.inputs = [["aa"], []]
        pr.inputDataPolicies = ["dl", "pr"]
        pr.events = [-1, -1]
        pr.bkQueries = ["Full", "fromPreviousProd"]
        pr.targets = ["Target1", "Target2"]
        pr.multicore = ["False", "True"]
        pr.processors = [(2, 4), (0, 0)]
        pr.outputModes = ["Local", "Any"]
        pr.ancestorDepths = [0, 1]
        pr._applyOptionalCorrections()
        prodsTypeListExpected = ["DataStripping", "Merge", "Merge", "Merge"]
        pluginsExpected = ["ByRun", "BySize", "BySize", "BySize"]
        stepsListDictExpected = [stepStripp, mergeStepBHADRON, mergeStepCALIBRA, mergeStepPIDMDST]
        stepsInProdExpected = [[1], [2], [3], [4]]
        outputSEsExpected = ["Tier1-DST", "Tier1-M-DST", "Tier1-M-DST", "Tier1-M-DST"]
        specialOutputSEsExpected = [{}, {}, {}, {}]
        prioritiesExpected = [1, 4, 4, 4]
        cpusExpected = [10, 100, 100, 100]
        bkQueriesExpected = ["Full", "fromPreviousProd", "fromPreviousProd", "fromPreviousProd"]
        groupSizesExpected = [1, 2, 2, 2]
        previousProdsExpected = [None, 1, 1, 1]
        removeInputsFlagExpected = [False, True, True, True]
        outputFileMasksExpected = ["", "dst", "dst", "dst"]
        outputFileStepsExpected = ["", "2", "2", "2"]
        inputsExpected = [["aa"], [], [], []]
        inputDataPoliciesExpected = ["dl", "pr", "pr", "pr"]
        eventsExpected = [-1, -1, -1, -1]
        targetsExpected = ["Target1", "Target2", "Target2", "Target2"]
        multicoreExpected = ["False", "True", "True", "True"]
        processorsExpected = [(2, 4), (0, 0), (0, 0), (0, 0)]
        outputModeExpected = ["Local", "Any", "Any", "Any"]
        ancestorDepthsExpected = [0, 1, 1, 1]
        self.assertEqual(pr.prodsTypeList, prodsTypeListExpected)
        self.assertEqual(pr.plugins, pluginsExpected)
        self.assertEqual(pr.stepsListDict, stepsListDictExpected)
        self.assertEqual(pr.stepsInProds, stepsInProdExpected)
        self.assertEqual(pr.outputSEs, outputSEsExpected)
        self.assertEqual(pr.specialOutputSEs, specialOutputSEsExpected)
        self.assertEqual(pr.priorities, prioritiesExpected)
        self.assertEqual(pr.cpus, cpusExpected)
        self.assertEqual(pr.groupSizes, groupSizesExpected)
        self.assertEqual(pr.bkQueries, bkQueriesExpected)
        self.assertEqual(pr.previousProds, previousProdsExpected)
        self.assertEqual(pr.removeInputsFlags, removeInputsFlagExpected)
        self.assertEqual(pr.outputFileMasks, outputFileMasksExpected)
        self.assertEqual(pr.outputFileSteps, outputFileStepsExpected)
        self.assertEqual(pr.inputs, inputsExpected)
        self.assertEqual(pr.inputDataPolicies, inputDataPoliciesExpected)
        self.assertEqual(pr.events, eventsExpected)
        self.assertEqual(pr.targets, targetsExpected)
        self.assertEqual(pr.multicore, multicoreExpected)
        self.assertEqual(pr.processors, processorsExpected)
        self.assertEqual(pr.outputModes, outputModeExpected)
        self.assertEqual(pr.ancestorDepths, ancestorDepthsExpected)

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.prodsTypeList = ["DataStripping", "Merge"]
        pr.plugins = ["ByRun", "ByRunFileTypeSizeWithFlush"]
        pr.stepsListDict = [stepStripp, mergeStep]
        pr.stepsList = [123, 456]
        pr.stepsInProds = [[1], [2]]
        pr.outputSEs = ["Tier1-DST", "Tier1-M-DST"]
        pr.specialOutputSEs = [{"Type1": "SE1"}, {}]
        pr.priorities = [1, 4]
        pr.cpus = [10, 100]
        pr.groupSizes = [1, 2]
        pr.bkFileType = ["Type1", "Type2"]
        pr.inputs = [[], ["aa"]]
        pr.removeInputsFlags = [False, True]
        pr.outputFileMasks = ["", "dst"]
        pr.outputFileSteps = ["", "2"]
        pr.inputDataPolicies = ["pr", "dl"]
        pr.events = [-1, -1]
        pr.bkQueries = ["Full", "fromPreviousProd"]
        pr.targets = ["Target1", "Target2"]
        pr.multicore = ["False", "True"]
        pr.processors = [2, 0]
        pr.outputModes = ["Local", "Any"]
        pr.ancestorDepths = [0, 0]
        pr._applyOptionalCorrections()
        prodsTypeListExpected = ["DataStripping", "Merge"]
        pluginsExpected = ["ByRun", "ByRunFileTypeSizeWithFlush"]
        stepsListDictExpected = [stepStripp, mergeStep]
        stepsInProdExpected = [[1], [2]]
        outputSEsExpected = ["Tier1-DST", "Tier1-M-DST"]
        specialOutputSEsExpected = [{"Type1": "SE1"}, {}]
        prioritiesExpected = [1, 4]
        cpusExpected = [10, 100]
        groupSizesExpected = [1, 2]
        bkQueriesExpected = ["Full", "fromPreviousProd"]
        previousProdsExpected = [None, 1]
        removeInputsFlagExpected = [False, True]
        outputFileMasksExpected = ["", "dst"]
        outputFileStepsExpected = ["", "2"]
        inputsExpected = [[], ["aa"]]
        inputDataPoliciesExpected = ["pr", "dl"]
        eventsExpected = [-1, -1]
        targetsExpected = ["Target1", "Target2"]
        multicoreExpected = ["False", "True"]
        processorsExpected = [2, 0]
        outputModeExpected = ["Local", "Any"]
        ancestorDepthsExpected = [0, 0]
        self.assertEqual(pr.prodsTypeList, prodsTypeListExpected)
        self.assertEqual(pr.plugins, pluginsExpected)
        self.assertEqual(pr.stepsListDict, stepsListDictExpected)
        self.assertEqual(pr.stepsInProds, stepsInProdExpected)
        self.assertEqual(pr.outputSEs, outputSEsExpected)
        self.assertEqual(pr.specialOutputSEs, specialOutputSEsExpected)
        self.assertEqual(pr.priorities, prioritiesExpected)
        self.assertEqual(pr.cpus, cpusExpected)
        self.assertEqual(pr.groupSizes, groupSizesExpected)
        self.assertEqual(pr.bkQueries, bkQueriesExpected)
        self.assertEqual(pr.previousProds, previousProdsExpected)
        self.assertEqual(pr.removeInputsFlags, removeInputsFlagExpected)
        self.assertEqual(pr.outputFileMasks, outputFileMasksExpected)
        self.assertEqual(pr.outputFileSteps, outputFileStepsExpected)
        self.assertEqual(pr.inputs, inputsExpected)
        self.assertEqual(pr.inputDataPolicies, inputDataPoliciesExpected)
        self.assertEqual(pr.events, eventsExpected)
        self.assertEqual(pr.targets, targetsExpected)
        self.assertEqual(pr.multicore, multicoreExpected)
        self.assertEqual(pr.processors, processorsExpected)
        self.assertEqual(pr.outputModes, outputModeExpected)
        self.assertEqual(pr.ancestorDepths, ancestorDepthsExpected)

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.prodsTypeList = ["Merge", "Merge"]
        pr.plugins = ["BySize", "ByRunFileTypeSizeWithFlush"]
        pr.stepsListDict = [mergeStep, mergeStep]
        pr.stepsList = [456, 456]
        pr.stepsInProds = [[1], [2]]
        pr.outputSEs = ["Tier1-DST", "Tier1-M-DST"]
        pr.specialOutputSEs = [{"Type1": "SE1"}, {}]
        pr.priorities = [1, 4]
        pr.cpus = [10, 100]
        pr.groupSizes = [1, 2]
        pr.removeInputsFlags = [True, True]
        pr.outputFileMasks = ["", "dst"]
        pr.outputFileSteps = ["", "2"]
        pr.inputs = [[], ["aa"]]
        pr.inputDataPolicies = ["dl", "pr"]
        pr.events = [-1, -1]
        pr.bkQueries = ["fromPreviousProd", "fromPreviousProd"]
        pr.targets = ["Target1", "Target2"]
        pr.multicore = ["False", "True"]
        pr.processors = [2, 0]
        pr.outputModes = ["Local", "Any"]
        pr.ancestorDepths = [1, 0]
        pr._applyOptionalCorrections()
        prodsTypeListExpected = ["Merge", "Merge", "Merge", "Merge"]
        pluginsExpected = ["BySize", "BySize", "BySize", "ByRunFileTypeSizeWithFlush"]
        stepsListDictExpected = [mergeStepBHADRON, mergeStepCALIBRA, mergeStepPIDMDST, mergeStep]
        stepsInProdExpected = [[1], [2], [3], [4]]
        outputSEsExpected = ["Tier1-DST", "Tier1-DST", "Tier1-DST", "Tier1-M-DST"]
        specialOutputSEsExpected = [{"Type1": "SE1"}, {"Type1": "SE1"}, {"Type1": "SE1"}, {}]
        prioritiesExpected = [1, 1, 1, 4]
        cpusExpected = [10, 10, 10, 100]
        groupSizesExpected = [1, 1, 1, 2]
        bkQueriesExpected = ["fromPreviousProd", "fromPreviousProd", "fromPreviousProd", "fromPreviousProd"]
        previousProdsExpected = [None, None, None, 1]
        removeInputsFlagExpected = [True, True, True, True]
        outputFileMasksExpected = ["", "", "", "dst"]
        outputFileStepsExpected = ["", "", "", "2"]
        inputsExpected = [[], [], [], ["aa"]]
        inputDataPoliciesExpected = ["dl", "dl", "dl", "pr"]
        eventsExpected = [-1, -1, -1, -1]
        targetsExpected = ["Target1", "Target1", "Target1", "Target2"]
        multicoreExpected = ["False", "False", "False", "True"]
        processorsExpected = [2, 2, 2, 0]
        outputModeExpected = ["Local", "Local", "Local", "Any"]
        ancestorDepthsExpected = [1, 1, 1, 0]
        self.assertEqual(pr.prodsTypeList, prodsTypeListExpected)
        self.assertEqual(pr.plugins, pluginsExpected)
        self.assertEqual(pr.stepsListDict, stepsListDictExpected)
        self.assertEqual(pr.stepsInProds, stepsInProdExpected)
        self.assertEqual(pr.outputSEs, outputSEsExpected)
        self.assertEqual(pr.specialOutputSEs, specialOutputSEsExpected)
        self.assertEqual(pr.priorities, prioritiesExpected)
        self.assertEqual(pr.cpus, cpusExpected)
        self.assertEqual(pr.groupSizes, groupSizesExpected)
        self.assertEqual(pr.bkQueries, bkQueriesExpected)
        self.assertEqual(pr.previousProds, previousProdsExpected)
        self.assertEqual(pr.removeInputsFlags, removeInputsFlagExpected)
        self.assertEqual(pr.outputFileMasks, outputFileMasksExpected)
        self.assertEqual(pr.outputFileSteps, outputFileStepsExpected)
        self.assertEqual(pr.inputs, inputsExpected)
        self.assertEqual(pr.inputDataPolicies, inputDataPoliciesExpected)
        self.assertEqual(pr.events, eventsExpected)
        self.assertEqual(pr.targets, targetsExpected)
        self.assertEqual(pr.multicore, multicoreExpected)
        self.assertEqual(pr.processors, processorsExpected)
        self.assertEqual(pr.outputModes, outputModeExpected)
        self.assertEqual(pr.ancestorDepths, ancestorDepthsExpected)

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.prodsTypeList = ["DataStripping"]
        pr.plugins = ["ByRun"]
        pr.stepsListDict = [stepStripp, mergeStep]
        pr.stepsList = [123]
        pr.stepsInProds = [[1, 2]]
        pr.priorities = [1]
        pr.outputSEs = ["Tier1-DST"]
        pr.cpus = [10]
        pr.groupSizes = [1]
        pr.bkFileType = "TYPE"
        pr.removeInputsFlags = [False]
        pr.outputFileMasks = [""]
        pr.outputFileSteps = ["2"]
        pr.inputs = [[]]
        pr.inputDataPolicies = [""]
        pr.events = [-1]
        pr.bkQueries = ["Full"]
        pr.targets = ["Target1"]
        pr.multicore = ["False"]
        pr.processors = [2]
        pr.outputModes = ["Local"]
        pr.ancestorDepths = [1]
        pr._applyOptionalCorrections()
        prodsTypeListExpected = ["DataStripping"]
        pluginsExpected = ["ByRun"]
        stepsListDictExpected = [stepStripp, mergeStep]
        stepsInProdExpected = [[1, 2]]
        outputSEsExpected = ["Tier1-DST"]
        prioritiesExpected = [1]
        cpusExpected = [10]
        groupSizesExpected = [1]
        bkQueriesExpected = ["Full"]
        previousProdsExpected = [None]
        removeInputsFlagExpected = [False]
        outputFileMasksExpected = [""]
        outputFileStepsExpected = ["2"]
        inputsExpected = [[]]
        inputDataPoliciesExpected = [""]
        eventsExpected = [-1]
        targetsExpected = ["Target1"]
        multicoreExpected = ["False"]
        processorsExpected = [2]
        outputModeExpected = ["Local"]
        ancestorDepthsExpected = [1]
        self.assertEqual(pr.prodsTypeList, prodsTypeListExpected)
        self.assertEqual(pr.plugins, pluginsExpected)
        self.assertEqual(pr.stepsListDict, stepsListDictExpected)
        self.assertEqual(pr.stepsInProds, stepsInProdExpected)
        self.assertEqual(pr.outputSEs, outputSEsExpected)
        self.assertEqual(pr.priorities, prioritiesExpected)
        self.assertEqual(pr.cpus, cpusExpected)
        self.assertEqual(pr.groupSizes, groupSizesExpected)
        self.assertEqual(pr.bkQueries, bkQueriesExpected)
        self.assertEqual(pr.previousProds, previousProdsExpected)
        self.assertEqual(pr.removeInputsFlags, removeInputsFlagExpected)
        self.assertEqual(pr.outputFileMasks, outputFileMasksExpected)
        self.assertEqual(pr.outputFileSteps, outputFileStepsExpected)
        self.assertEqual(pr.inputs, inputsExpected)
        self.assertEqual(pr.inputDataPolicies, inputDataPoliciesExpected)
        self.assertEqual(pr.events, eventsExpected)
        self.assertEqual(pr.targets, targetsExpected)
        self.assertEqual(pr.multicore, multicoreExpected)
        self.assertEqual(pr.processors, processorsExpected)
        self.assertEqual(pr.outputModes, outputModeExpected)
        self.assertEqual(pr.ancestorDepths, ancestorDepthsExpected)

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.prodsTypeList = ["DataStripping", "Merge", "Merge"]
        pr.plugins = ["ByRun", "BySize", "ByRunFileTypeSizeWithFlush"]
        pr.stepsListDict = [stepStripp, mergeStep, mergeStep]
        pr.stepsList = [123, 456, 456]
        pr.stepsInProds = [[1], [2], [3]]
        pr.outputSEs = ["Tier1-DST", "Tier1-M-DST", "Tier1-M-DST"]
        pr.priorities = [1, 4, 5]
        pr.cpus = [10, 100, 1000]
        pr.groupSizes = [1, 2, 3]
        pr.bkFileType = "TYPE"
        pr.removeInputsFlags = [False, True, True]
        pr.outputFileMasks = ["", "dst", ""]
        pr.outputFileSteps = ["", "2", ""]
        pr.inputs = [[], ["aa"], ["bb"]]
        pr.inputDataPolicies = ["", "dl", "pr"]
        pr.bkQueries = ["Full", "fromPreviousProd", "fromPreviousProd"]
        pr.events = [-1, -1, -1]
        pr.targets = ["Target1", "Target2", "Target3"]
        pr.multicore = ["False", "True", "False"]
        pr.processors = [2, 0, 4]
        pr.outputModes = ["Local", "Any", "Any"]
        pr.ancestorDepths = [1, 2, 3]
        pr._applyOptionalCorrections()
        prodsTypeListExpected = ["DataStripping", "Merge", "Merge", "Merge", "Merge"]
        pluginsExpected = ["ByRun", "BySize", "BySize", "BySize", "ByRunFileTypeSizeWithFlush"]
        stepsListDictExpected = [stepStripp, mergeStepBHADRON, mergeStepCALIBRA, mergeStepPIDMDST, mergeStep]
        stepsInProdExpected = [[1], [2], [3], [4], [5]]
        outputSEsExpected = ["Tier1-DST", "Tier1-M-DST", "Tier1-M-DST", "Tier1-M-DST", "Tier1-M-DST"]
        prioritiesExpected = [1, 4, 4, 4, 5]
        groupSizesExpected = [1, 2, 2, 2, 3]
        cpusExpected = [10, 100, 100, 100, 1000]
        bkQueriesExpected = ["Full", "fromPreviousProd", "fromPreviousProd", "fromPreviousProd", "fromPreviousProd"]
        previousProdsExpected = [None, 1, 1, 1, 2]
        removeInputsFlagExpected = [False, True, True, True, True]
        outputFileMasksExpected = ["", "dst", "dst", "dst", ""]
        outputFileStepsExpected = ["", "2", "2", "2", ""]
        inputsExpected = [[], ["aa"], ["aa"], ["aa"], ["bb"]]
        inputDataPoliciesExpected = ["", "dl", "dl", "dl", "pr"]
        eventsExpected = [-1, -1, -1, -1, -1]
        targetsExpected = ["Target1", "Target2", "Target2", "Target2", "Target3"]
        multicoreExpected = ["False", "True", "True", "True", "False"]
        processorsExpected = [2, 0, 0, 0, 4]
        outputModeExpected = ["Local", "Any", "Any", "Any", "Any"]
        ancestorDepthsExpected = [1, 2, 2, 2, 3]
        self.assertEqual(pr.prodsTypeList, prodsTypeListExpected)
        self.assertEqual(pr.plugins, pluginsExpected)
        self.assertEqual(pr.stepsListDict, stepsListDictExpected)
        self.assertEqual(pr.stepsInProds, stepsInProdExpected)
        self.assertEqual(pr.outputSEs, outputSEsExpected)
        self.assertEqual(pr.priorities, prioritiesExpected)
        self.assertEqual(pr.cpus, cpusExpected)
        self.assertEqual(pr.groupSizes, groupSizesExpected)
        self.assertEqual(pr.bkQueries, bkQueriesExpected)
        self.assertEqual(pr.previousProds, previousProdsExpected)
        self.assertEqual(pr.removeInputsFlags, removeInputsFlagExpected)
        self.assertEqual(pr.outputFileMasks, outputFileMasksExpected)
        self.assertEqual(pr.outputFileSteps, outputFileStepsExpected)
        self.assertEqual(pr.inputs, inputsExpected)
        self.assertEqual(pr.inputDataPolicies, inputDataPoliciesExpected)
        self.assertEqual(pr.events, eventsExpected)
        self.assertEqual(pr.targets, targetsExpected)
        self.assertEqual(pr.multicore, multicoreExpected)
        self.assertEqual(pr.processors, processorsExpected)
        self.assertEqual(pr.outputModes, outputModeExpected)
        self.assertEqual(pr.ancestorDepths, ancestorDepthsExpected)

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.prodsTypeList = ["DataStripping", "Merge"]
        pr.plugins = ["ByRun", "BySize"]
        pr.stepsListDict = [stepStripp, mergeStep]
        pr.stepsList = [123, 456]
        pr.stepsInProds = [[3], [4]]
        pr.outputSEs = ["Tier1-DST", "Tier1-M-DST"]
        pr.priorities = [1, 4]
        pr.cpus = [10, 100]
        pr.groupSizes = [1, 2]
        pr.bkFileType = "TYPE"
        pr.removeInputsFlags = [False, True]
        pr.outputFileMasks = ["", "dst"]
        pr.outputFileSteps = ["", "2"]
        pr.inputs = [[], []]
        pr.inputDataPolicies = ["", "dl"]
        pr.events = [-1, -1]
        pr.bkQueries = ["Full", "fromPreviousProd"]
        pr.targets = ["Target1", "Target2"]
        pr.multicore = ["False", "True"]
        pr.processors = [2, 0]
        pr.outputModes = ["Local", "Any"]
        pr.ancestorDepths = [1, 0]
        pr._applyOptionalCorrections()
        prodsTypeListExpected = ["DataStripping", "Merge", "Merge", "Merge"]
        pluginsExpected = ["ByRun", "BySize", "BySize", "BySize"]
        stepsListDictExpected = [stepStripp, mergeStepBHADRON, mergeStepCALIBRA, mergeStepPIDMDST]
        stepsInProdExpected = [[3], [4], [5], [6]]
        outputSEsExpected = ["Tier1-DST", "Tier1-M-DST", "Tier1-M-DST", "Tier1-M-DST"]
        prioritiesExpected = [1, 4, 4, 4]
        cpusExpected = [10, 100, 100, 100]
        groupSizesExpected = [1, 2, 2, 2]
        bkQueriesExpected = ["Full", "fromPreviousProd", "fromPreviousProd", "fromPreviousProd"]
        previousProdsExpected = [None, 1, 1, 1]
        removeInputsFlagExpected = [False, True, True, True]
        outputFileMasksExpected = ["", "dst", "dst", "dst"]
        outputFileStepsExpected = ["", "2", "2", "2"]
        inputsExpected = [[], [], [], []]
        inputDataPoliciesExpected = ["", "dl", "dl", "dl"]
        eventsExpected = [-1, -1, -1, -1]
        targetsExpected = ["Target1", "Target2", "Target2", "Target2"]
        multicoreExpected = ["False", "True", "True", "True"]
        processorsExpected = [2, 0, 0, 0]
        outputModeExpected = ["Local", "Any", "Any", "Any"]
        ancestorDepthsExpected = [1, 0, 0, 0]
        self.assertEqual(pr.prodsTypeList, prodsTypeListExpected)
        self.assertEqual(pr.plugins, pluginsExpected)
        self.assertEqual(pr.stepsListDict, stepsListDictExpected)
        self.assertEqual(pr.stepsInProds, stepsInProdExpected)
        self.assertEqual(pr.outputSEs, outputSEsExpected)
        self.assertEqual(pr.priorities, prioritiesExpected)
        self.assertEqual(pr.cpus, cpusExpected)
        self.assertEqual(pr.groupSizes, groupSizesExpected)
        self.assertEqual(pr.bkQueries, bkQueriesExpected)
        self.assertEqual(pr.previousProds, previousProdsExpected)
        self.assertEqual(pr.removeInputsFlags, removeInputsFlagExpected)
        self.assertEqual(pr.outputFileMasks, outputFileMasksExpected)
        self.assertEqual(pr.outputFileSteps, outputFileStepsExpected)
        self.assertEqual(pr.inputs, inputsExpected)
        self.assertEqual(pr.inputDataPolicies, inputDataPoliciesExpected)
        self.assertEqual(pr.events, eventsExpected)
        self.assertEqual(pr.targets, targetsExpected)
        self.assertEqual(pr.multicore, multicoreExpected)
        self.assertEqual(pr.processors, processorsExpected)
        self.assertEqual(pr.outputModes, outputModeExpected)
        self.assertEqual(pr.ancestorDepths, ancestorDepthsExpected)

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.prodsTypeList = ["DataStripping"]
        pr.plugins = ["ByRun"]
        pr.stepsListDict = [stepStripp, mergeStep]
        pr.stepsList = [123, 456]
        pr.stepsInProds = [[3, 4]]
        pr.priorities = [1]
        pr.outputSEs = ["Tier1-DST"]
        pr.cpus = [10]
        pr.groupSizes = [1]
        pr.removeInputsFlags = [False]
        pr.outputFileMasks = ["dst"]
        pr.outputFileSteps = ["2"]
        pr.bkFileType = "TYPE"
        pr.inputs = [[]]
        pr.inputDataPolicies = ["dl"]
        pr.events = [-1]
        pr.bkQueries = ["Full"]
        pr.targets = ["Target1"]
        pr.multicore = ["False"]
        pr.processors = [2]
        pr.outputModes = ["Local"]
        pr.ancestorDepths = [1]
        pr._applyOptionalCorrections()
        prodsTypeListExpected = ["DataStripping"]
        pluginsExpected = ["ByRun"]
        stepsListDictExpected = [stepStripp, mergeStep]
        stepsInProdExpected = [[3, 4]]
        outputSEsExpected = ["Tier1-DST"]
        prioritiesExpected = [1]
        cpusExpected = [10]
        groupSizesExpected = [1]
        bkQueriesExpected = ["Full"]
        previousProdsExpected = [None]
        removeInputsFlagExpected = [False]
        outputFileMasksExpected = ["dst"]
        outputFileStepsExpected = ["2"]
        inputsExpected = [[]]
        inputDataPoliciesExpected = ["dl"]
        eventsExpected = [-1]
        targetsExpected = ["Target1"]
        multicoreExpected = ["False"]
        processorsExpected = [2]
        outputModeExpected = ["Local"]
        ancestorDepthsExpected = [1]
        self.assertEqual(pr.prodsTypeList, prodsTypeListExpected)
        self.assertEqual(pr.plugins, pluginsExpected)
        self.assertEqual(pr.stepsListDict, stepsListDictExpected)
        self.assertEqual(pr.stepsInProds, stepsInProdExpected)
        self.assertEqual(pr.outputSEs, outputSEsExpected)
        self.assertEqual(pr.priorities, prioritiesExpected)
        self.assertEqual(pr.cpus, cpusExpected)
        self.assertEqual(pr.groupSizes, groupSizesExpected)
        self.assertEqual(pr.bkQueries, bkQueriesExpected)
        self.assertEqual(pr.previousProds, previousProdsExpected)
        self.assertEqual(pr.removeInputsFlags, removeInputsFlagExpected)
        self.assertEqual(pr.outputFileMasks, outputFileMasksExpected)
        self.assertEqual(pr.outputFileSteps, outputFileStepsExpected)
        self.assertEqual(pr.inputs, inputsExpected)
        self.assertEqual(pr.inputDataPolicies, inputDataPoliciesExpected)
        self.assertEqual(pr.events, eventsExpected)
        self.assertEqual(pr.targets, targetsExpected)
        self.assertEqual(pr.multicore, multicoreExpected)
        self.assertEqual(pr.processors, processorsExpected)
        self.assertEqual(pr.outputModes, outputModeExpected)
        self.assertEqual(pr.ancestorDepths, ancestorDepthsExpected)

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.prodsTypeList = ["Merge", "Merge"]
        pr.plugins = ["BySize", "ByRunFileTypeSizeWithFlush"]
        pr.stepsListDict = [mergeStep, mergeStep]
        pr.stepsList = [456, 456]
        pr.stepsInProds = [[3], [4]]
        pr.outputSEs = ["Tier1-DST", "Tier1-M-DST"]
        pr.priorities = [1, 4]
        pr.cpus = [10, 100]
        pr.groupSizes = [1, 2]
        pr.removeInputsFlags = [True, True]
        pr.outputFileMasks = ["", "dst"]
        pr.outputFileSteps = ["", "2"]
        pr.inputs = [["aa"], ["bb"]]
        pr.inputDataPolicies = ["dl", "pr"]
        pr.events = [-1, -1]
        pr.bkQueries = ["Full", "fromPreviousProd"]
        pr.targets = ["Target1", "Target2"]
        pr.multicore = ["False", "True"]
        pr.processors = [2, 0]
        pr.outputModes = ["Local", "Any"]
        pr.ancestorDepths = [1, 0]
        pr._applyOptionalCorrections()
        prodsTypeListExpected = ["Merge", "Merge", "Merge", "Merge"]
        pluginsExpected = ["BySize", "BySize", "BySize", "ByRunFileTypeSizeWithFlush"]
        stepsListDictExpected = [mergeStepBHADRON, mergeStepCALIBRA, mergeStepPIDMDST, mergeStep]
        stepsInProdExpected = [[3], [4], [5], [6]]
        outputSEsExpected = ["Tier1-DST", "Tier1-DST", "Tier1-DST", "Tier1-M-DST"]
        prioritiesExpected = [1, 1, 1, 4]
        cpusExpected = [10, 10, 10, 100]
        groupSizesExpected = [1, 1, 1, 2]
        bkQueriesExpected = ["Full", "Full", "Full", "fromPreviousProd"]
        previousProdsExpected = [None, None, None, 1]
        removeInputsFlagExpected = [True, True, True, True]
        outputFileMasksExpected = ["", "", "", "dst"]
        outputFileStepsExpected = ["", "", "", "2"]
        inputsExpected = [["aa"], ["aa"], ["aa"], ["bb"]]
        inputDataPoliciesExpected = ["dl", "dl", "dl", "pr"]
        eventsExpected = [-1, -1, -1, -1]
        targetsExpected = ["Target1", "Target1", "Target1", "Target2"]
        multicoreExpected = ["False", "False", "False", "True"]
        processorsExpected = [2, 2, 2, 0]
        outputModeExpected = ["Local", "Local", "Local", "Any"]
        ancestorDepthsExpected = [1, 1, 1, 0]
        self.assertEqual(pr.prodsTypeList, prodsTypeListExpected)
        self.assertEqual(pr.plugins, pluginsExpected)
        self.assertEqual(pr.stepsListDict, stepsListDictExpected)
        self.assertEqual(pr.stepsInProds, stepsInProdExpected)
        self.assertEqual(pr.outputSEs, outputSEsExpected)
        self.assertEqual(pr.priorities, prioritiesExpected)
        self.assertEqual(pr.cpus, cpusExpected)
        self.assertEqual(pr.groupSizes, groupSizesExpected)
        self.assertEqual(pr.bkQueries, bkQueriesExpected)
        self.assertEqual(pr.previousProds, previousProdsExpected)
        self.assertEqual(pr.removeInputsFlags, removeInputsFlagExpected)
        self.assertEqual(pr.outputFileMasks, outputFileMasksExpected)
        self.assertEqual(pr.outputFileSteps, outputFileStepsExpected)
        self.assertEqual(pr.inputs, inputsExpected)
        self.assertEqual(pr.inputDataPolicies, inputDataPoliciesExpected)
        self.assertEqual(pr.events, eventsExpected)
        self.assertEqual(pr.targets, targetsExpected)
        self.assertEqual(pr.multicore, multicoreExpected)
        self.assertEqual(pr.processors, processorsExpected)
        self.assertEqual(pr.outputModes, outputModeExpected)
        self.assertEqual(pr.ancestorDepths, ancestorDepthsExpected)

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.prodsTypeList = ["MCSimulation", "Merge"]
        pr.plugins = ["", "ByRunFileTypeSizeWithFlush"]
        pr.stepsListDict = [stepMC, mergeStep]
        pr.stepsList = [123, 456]
        pr.stepsInProds = [[1], [2]]
        pr.outputSEs = ["Tier1-DST", "Tier1-M-DST"]
        pr.priorities = [1, 4]
        pr.cpus = [10, 100]
        pr.groupSizes = [1, 2]
        pr.removeInputsFlags = [False, True]
        pr.outputFileMasks = ["", "dst"]
        pr.outputFileSteps = ["", "2"]
        pr.inputs = [[], []]
        pr.inputDataPolicies = ["", "dl"]
        pr.events = [100, -1]
        pr.bkQueries = ["", "fromPreviousProd"]
        pr.targets = ["Target1", "Target2"]
        pr.multicore = ["False", "True"]
        pr.processors = [2, 0]
        pr.outputModes = ["Local", "Any"]
        pr.ancestorDepths = [1, 0]
        pr._applyOptionalCorrections()
        prodsTypeListExpected = ["MCSimulation", "Merge"]
        pluginsExpected = ["", "ByRunFileTypeSizeWithFlush"]
        stepsListDictExpected = [stepMC, mergeStep]
        stepsInProdExpected = [[1], [2]]
        outputSEsExpected = ["Tier1-DST", "Tier1-M-DST"]
        prioritiesExpected = [1, 4]
        cpusExpected = [10, 100]
        groupSizesExpected = [1, 2]
        bkQueriesExpected = ["", "fromPreviousProd"]
        previousProdsExpected = [None, 1]
        removeInputsFlagExpected = [False, True]
        outputFileMasksExpected = ["", "dst"]
        outputFileStepsExpected = ["", "2"]
        inputsExpected = [[], []]
        inputDataPoliciesExpected = ["", "dl"]
        eventsExpected = [100, -1]
        targetsExpected = ["Target1", "Target2"]
        multicoreExpected = ["False", "True"]
        processorsExpected = [2, 0]
        outputModeExpected = ["Local", "Any"]
        ancestorDepthsExpected = [1, 0]
        self.assertEqual(pr.prodsTypeList, prodsTypeListExpected)
        self.assertEqual(pr.plugins, pluginsExpected)
        self.assertEqual(pr.stepsListDict, stepsListDictExpected)
        self.assertEqual(pr.stepsInProds, stepsInProdExpected)
        self.assertEqual(pr.outputSEs, outputSEsExpected)
        self.assertEqual(pr.priorities, prioritiesExpected)
        self.assertEqual(pr.cpus, cpusExpected)
        self.assertEqual(pr.groupSizes, groupSizesExpected)
        self.assertEqual(pr.bkQueries, bkQueriesExpected)
        self.assertEqual(pr.previousProds, previousProdsExpected)
        self.assertEqual(pr.removeInputsFlags, removeInputsFlagExpected)
        self.assertEqual(pr.outputFileMasks, outputFileMasksExpected)
        self.assertEqual(pr.outputFileSteps, outputFileStepsExpected)
        self.assertEqual(pr.inputs, inputsExpected)
        self.assertEqual(pr.inputDataPolicies, inputDataPoliciesExpected)
        self.assertEqual(pr.events, eventsExpected)
        self.assertEqual(pr.targets, targetsExpected)
        self.assertEqual(pr.multicore, multicoreExpected)
        self.assertEqual(pr.processors, processorsExpected)
        self.assertEqual(pr.outputModes, outputModeExpected)
        self.assertEqual(pr.ancestorDepths, ancestorDepthsExpected)

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.prodsTypeList = ["MCSimulation", "Merge", "Merge"]
        pr.plugins = ["", "ByRunFileTypeSizeWithFlush", "RootMerging"]
        pr.stepsListDict = [stepMC, mergeStep, stepHistoMergingDict]
        pr.stepsList = [123, 456, 789]
        pr.stepsInProds = [[1], [2], [3]]
        pr.outputSEs = ["Tier1-DST", "Tier1-M-DST", "Tier1-M-DST"]
        pr.priorities = [1, 4, 5]
        pr.cpus = [10, 100, 100]
        pr.groupSizes = [1, 2, 3]
        pr.removeInputsFlags = [False, True, True]
        pr.outputFileMasks = ["", "dst", "root"]
        pr.outputFileSteps = ["", "2", "3"]
        pr.inputs = [[], [], []]
        pr.inputDataPolicies = ["", "dl", "dl"]
        pr.events = [100, -1, -1]
        pr.bkQueries = ["", "fromPreviousProd", "fromPreviousProd"]
        pr.targets = ["Target1", "Target2", "T3"]
        pr.multicore = ["False", "True", "True"]
        pr.processors = [2, 0, 0]
        pr.outputModes = ["Local", "Any", "Any"]
        pr.ancestorDepths = [1, 0, 0]
        pr._applyOptionalCorrections()
        prodsTypeListExpected = ["MCSimulation", "Merge", "Merge"]
        pluginsExpected = ["", "ByRunFileTypeSizeWithFlush", "RootMerging"]
        stepsListDictExpected = [stepMC, mergeStep, stepHistoMergingDict]
        stepsInProdExpected = [[1], [2], [3]]
        outputSEsExpected = ["Tier1-DST", "Tier1-M-DST", "Tier1-M-DST"]
        prioritiesExpected = [1, 4, 5]
        cpusExpected = [10, 100, 100]
        groupSizesExpected = [1, 2, 3]
        bkQueriesExpected = ["", "fromPreviousProd", "fromPreviousProd"]
        previousProdsExpected = [None, 1, 2]
        removeInputsFlagExpected = [False, True, True]
        outputFileMasksExpected = ["", "dst", "root"]
        outputFileStepsExpected = ["", "2", "3"]
        inputsExpected = [[], [], []]
        inputDataPoliciesExpected = ["", "dl", "dl"]
        eventsExpected = [100, -1, -1]
        targetsExpected = ["Target1", "Target2", "T3"]
        multicoreExpected = ["False", "True", "True"]
        processorsExpected = [2, 0, 0]
        outputModeExpected = ["Local", "Any", "Any"]
        ancestorDepthsExpected = [1, 0, 0]
        self.assertEqual(pr.prodsTypeList, prodsTypeListExpected)
        self.assertEqual(pr.plugins, pluginsExpected)
        self.assertEqual(pr.stepsListDict, stepsListDictExpected)
        self.assertEqual(pr.stepsInProds, stepsInProdExpected)
        self.assertEqual(pr.outputSEs, outputSEsExpected)
        self.assertEqual(pr.priorities, prioritiesExpected)
        self.assertEqual(pr.cpus, cpusExpected)
        self.assertEqual(pr.groupSizes, groupSizesExpected)
        self.assertEqual(pr.bkQueries, bkQueriesExpected)
        self.assertEqual(pr.previousProds, previousProdsExpected)
        self.assertEqual(pr.removeInputsFlags, removeInputsFlagExpected)
        self.assertEqual(pr.outputFileMasks, outputFileMasksExpected)
        self.assertEqual(pr.outputFileSteps, outputFileStepsExpected)
        self.assertEqual(pr.inputs, inputsExpected)
        self.assertEqual(pr.inputDataPolicies, inputDataPoliciesExpected)
        self.assertEqual(pr.events, eventsExpected)
        self.assertEqual(pr.targets, targetsExpected)
        self.assertEqual(pr.multicore, multicoreExpected)
        self.assertEqual(pr.processors, processorsExpected)
        self.assertEqual(pr.outputModes, outputModeExpected)
        self.assertEqual(pr.ancestorDepths, ancestorDepthsExpected)

    def test_getProdsDescriptionDict(self):
        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.stepsList = [123, 456, 456, 456, 789]
        pr.prodsTypeList = ["DataStripping", "Merge", "Merge", "Merge"]
        pr.removeInputsFlags = [False, True, True, True]
        pr.inputs = [[], [], [], []]
        pr.targets = ["", "", "", ""]
        pr.groupSizes = [1, 1, 2, 3]
        pr.inputDataPolicies = ["download", "download", "download", "download"]
        pr.outputFileMasks = ["", "", "", ""]
        pr.outputFileSteps = ["", "", "", ""]
        pr.stepsInProds = [[1, 2], [3], [4], [5]]
        pr.bkQueries = ["Full", "fromPreviousProd", "fromPreviousProd", "fromPreviousProd"]
        #     pr.outputSEs = ['Tier1-Buffer', 'Tier1-DST', 'Tier1-DST']
        #     pr.specialOutputSEs = [{}, {}, {}]
        pr.outputSEsPerFileType = [{"T1": "SE1"}, {"T1": "SE1"}, {"T1": "SE1"}, {"T1": "SE1"}]
        pr.priorities = [5, 8, 9, 9]
        pr.cpus = [1000000, 300000, 10000, 1000]
        pr.plugins = ["ByRun", "BySize", "BySize", "BySize"]
        pr.previousProds = [None, 1, 1, 1]
        pr.events = [-1, -1, -1, -1]
        pr.multicore = ["False", "False", "True", "True"]
        pr.processors = [(4, 4), (2, 2), (0, 2), (0, 0)]
        pr.outputModes = ["Any", "Local", "Any", "Any"]
        pr.ancestorDepths = [0, 0, 0, 0]

        pr.stepsListDict = [
            {
                "StepId": 123,
                "StepName": "Stripping14-Stripping",
                "ApplicationName": "DaVinci",
                "ApplicationVersion": "v2r2",
                "OptionFiles": "optsFiles",
                "Visible": "Yes",
                "ExtraPackages": "eps",
                "ProcessingPass": "procPass",
                "OptionsFormat": "",
                "DDDB": "",
                "CONDDB": "123456",
                "DQTag": "",
                "isMulticore": "False",
                "fileTypesIn": ["SDST"],
                "fileTypesOut": ["BHADRON.DST", "CALIBRATION.DST"],
            },
            {
                "StepId": 456,
                "StepName": "Merge",
                "ApplicationName": "LHCb",
                "ApplicationVersion": "v1r2",
                "OptionFiles": "optsFiles",
                "Visible": "Yes",
                "ExtraPackages": "eps",
                "ProcessingPass": "procPass",
                "OptionsFormat": "",
                "DDDB": "",
                "CONDDB": "123456",
                "DQTag": "",
                "isMulticore": "False",
                "prodStepID": "456['BHADRON.DST']",
                "fileTypesIn": ["BHADRON.DST"],
                "fileTypesOut": ["BHADRON.DST"],
            },
            {
                "StepId": 456,
                "StepName": "Merge",
                "ApplicationName": "LHCb",
                "ApplicationVersion": "v1r2",
                "OptionFiles": "optsFiles",
                "Visible": "Yes",
                "ExtraPackages": "eps",
                "ProcessingPass": "procPass",
                "OptionsFormat": "",
                "DDDB": "",
                "CONDDB": "123456",
                "DQTag": "",
                "isMulticore": "False",
                "prodStepID": "456['CALIBRATION.DST']",
                "fileTypesIn": ["CALIBRATION.DST"],
                "fileTypesOut": ["CALIBRATION.DST"],
            },
            {
                "StepId": 456,
                "StepName": "Merge",
                "ApplicationName": "LHCb",
                "ApplicationVersion": "v1r2",
                "OptionFiles": "optsFiles",
                "Visible": "Yes",
                "ExtraPackages": "eps",
                "ProcessingPass": "procPass",
                "OptionsFormat": "",
                "DDDB": "",
                "CONDDB": "123456",
                "DQTag": "",
                "isMulticore": "False",
                "prodStepID": "456['PID.MDST']",
                "fileTypesIn": ["PID.MDST"],
                "fileTypesOut": ["PID.MDST"],
            },
            {
                "StepId": 789,
                "StepName": "Merge",
                "ApplicationName": "Noether",
                "ApplicationVersion": "v1r2",
                "OptionFiles": "optsFiles",
                "Visible": "Yes",
                "ExtraPackages": "eps",
                "ProcessingPass": "procPass",
                "OptionsFormat": "",
                "DDDB": "",
                "CONDDB": "123456",
                "DQTag": "",
                "isMulticore": "False",
                "prodStepID": "789['DAVINCIHIST', 'BRUNELHIST']",
                "fileTypesIn": ["DAVINCIHIST", "BRUNELHIST"],
                "fileTypesOut": ["ROOT"],
            },
        ]
        res = pr._getProdsDescriptionDict()

        self.maxDiff = None
        self.assertEqual(res, prodsDict)

    def test__addStepsToProd(self):
        stepsInProd = [
            {
                "StepId": 123,
                "StepName": "Stripping14-Stripping",
                "ApplicationName": "DaVinci",
                "ApplicationVersion": "v2r2",
                "ExtraOptions": "",
                "EventTimeout": "",
                "OptionFiles": "optsFiles",
                "Visible": "Yes",
                "ExtraPackages": "eps",
                "ProcessingPass": "procPass",
                "OptionsFormat": "",
                "SystemConfig": "",
                "DDDB": "",
                "CONDDB": "123456",
                "DQTag": "",
                "isMulticore": "N",
                "prodStepID": "123['SDST']",
                "mcTCK": "",
                "visibilityFlag": [
                    {"Visibility": "N", "FileType": "BHADRON.DST"},
                    {"Visibility": "N", "FileType": "CALIBRATION.DST"},
                ],
                "fileTypesIn": ["SDST"],
                "fileTypesOut": ["BHADRON.DST", "CALIBRATION.DST"],
            },
            {
                "StepId": 456,
                "StepName": "Merge",
                "ApplicationName": "LHCb",
                "ApplicationVersion": "v1r2",
                "ExtraOptions": "",
                "EventTimeout": "",
                "OptionFiles": "optsFiles",
                "Visible": "Yes",
                "ExtraPackages": "eps",
                "ProcessingPass": "procPass",
                "OptionsFormat": "",
                "SystemConfig": "x86_64+avx2-centos7-gcc7-opt",
                "prodStepID": "456['BHADRON.DST', 'CALIBRATION.DST']",
                "DDDB": "",
                "CONDDB": "123456",
                "DQTag": "",
                "isMulticore": "N",
                "mcTCK": "",
                "visibilityFlag": [
                    {"Visibility": "Y", "FileType": "BHADRON.DST"},
                    {"Visibility": "Y", "FileType": "CALIBRATION.DST"},
                ],
                "fileTypesIn": ["BHADRON.DST", "CALIBRATION.DST"],
                "fileTypesOut": ["BHADRON.DST", "CALIBRATION.DST"],
            },
        ]

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        prod = Production()
        prod = pr._addStepsToProd(prod, stepsInProd, True)
        self.assertEqual(prod.gaudiSteps, ["DaVinci_1", "LHCb_2"])

    # def test__getStepsInProdDAG( self ):

    #   pr = ProductionRequest( self.bkClientFake, self.diracProdIn )
    #   # stepsListDict = {}
    #   # for prodDict in prodsDict.values():
    #   #   stepsInProdDAG = pr._getStepsInProdDAG(prodDict, stepsListDict)
    #   #   self.assertEqual(stepsInProdDAG.graph, {})

    #   stepsListDict = [step1Dict, mergeStepBHADRON]
    #   stepsInProdDAG = pr._getStepsInProdDAG(prodsDict[1], stepsListDict)

    #   self.assertEqual(
    #       stepsInProdDAG.graph,
    #       {frozenset([('DDDB', ''),
    #                   ('ExtraPackages', 'eps'),
    #                   ('OptionsFormat', ''),
    #                   ('visibilityFlag', frozenset([
    #                       {'Visible':'N', 'FileType': 'BHADRON.DST'},
    #                       {'Visible':'N', 'FileType': 'CALIBRATION.DST'},
    #                   ])),
    #                   ('fileTypesOut', frozenset(['CALIBRATION.DST', 'BHADRON.DST'])),
    #                   ('DQTag', ''),
    #                   ('CONDDB', '123456'),
    #                   ('mcTCK', ''),
    #                   ('ApplicationName', 'DaVinci'),
    #                   ('ApplicationVersion', 'v2r2'),
    #                   ('Visible', 'Yes'),
    #                   ('ProcessingPass', 'procPass'),
    #                   ('fileTypesIn', frozenset(['SDST'])),
    #                   ('isMulticore', 'N'),
    #                   ('SystemConfig', ''),
    #                   ('prodStepID', "123['SDST']"),
    #                   ('StepName', 'Stripping14-Stripping'),
    #                   ('StepId', 123),
    #                   ('OptionFiles', 'optsFiles'),
    #                   ('ExtraOptions', '')]): set([frozenset([
    #                       ('DDDB', 'head-20110302'),
    #                       ('ExtraPackages', 'AppConfig.v3r104'),
    #                       ('fileTypesOut', frozenset(['BHADRON.DST'])),
    #                       ('CONDDB', 'head-20110407'),
    #                       ('mcTCK', ''),
    #                       ('visibilityFlag', frozenset([''])),
    #                       ('ApplicationName', 'DaVinci'),
    #                       ('ApplicationVersion', 'v28r3p1'),
    #                       ('ProcessingPass', 'Merging'),
    #                       ('fileTypesIn', frozenset(['BHADRON.DST'])),
    #                       ('prodStepID', "456['BHADRON.DST']"),
    #                       ('StepName', 'Stripping14-Merging'),
    #                       ('StepId', 456),
    #                       ('SystemConfig', ''),
    #                       ('Usable', 'Yes'),
    #                       ('Visible', 'N'),
    #                       ('OptionFiles', '$APPCONFIGOPTS/Merging/DV-Stripping14-Merging.py')
    #                   ])]),
    #         frozenset([('DDDB', 'head-20110302'),
    #                   ('ExtraPackages', 'AppConfig.v3r104'),
    #                   ('fileTypesOut', frozenset(['BHADRON.DST'])),
    #                   ('CONDDB', 'head-20110407'),
    #                   ('mcTCK', ''),
    #                   ('visibilityFlag', frozenset([''])),
    #                   ('ApplicationName', 'DaVinci'),
    #                   ('ApplicationVersion', 'v28r3p1'),
    #                   ('ProcessingPass', 'Merging'),
    #                   ('fileTypesIn', frozenset(['BHADRON.DST'])),
    #                   ('prodStepID', "456['BHADRON.DST']"),
    #                   ('StepName', 'Stripping14-Merging'),
    #                   ('StepId', 456),
    #                   ('SystemConfig', ''),
    #                   ('Usable', 'Yes'),
    #                   ('Visible', 'N'),
    #                   ('OptionFiles', '$APPCONFIGOPTS/Merging/DV-Stripping14-Merging.py')]): set([]),
    #       }
    #     )

    def test__getBKKQuery(self):
        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.dataTakingConditions = "dataTC"
        pr.processingPass = "procePass"
        pr.dqFlag = "OK,AA, BB"
        pr.startRun = "123"
        pr.endRun = "456"
        res = pr._getBKKQuery()

        resExpected = {
            "DataTakingConditions": "dataTC",
            "ProcessingPass": "procePass",
            "FileType": "",
            "EventType": "",
            "ConfigName": "test",
            "ConfigVersion": "certification",
            "DataQualityFlag": "OK;;;AA;;;BB",
            "StartRun": 123,
            "EndRun": 456,
        }
        self.assertEqual(res, resExpected)

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.dataTakingConditions = "dataTC"
        pr.processingPass = "procePass"
        pr.dqFlag = "OK,AA, BB"
        pr.bkFileType = ["type1", "type2"]
        pr.startRun = "123"
        pr.endRun = "456"
        res = pr._getBKKQuery()
        resExpected = {
            "DataTakingConditions": "dataTC",
            "ProcessingPass": "procePass",
            "FileType": "type1;;;type2",
            "EventType": "",
            "ConfigName": "test",
            "ConfigVersion": "certification",
            "DataQualityFlag": "OK;;;AA;;;BB",
            "StartRun": 123,
            "EndRun": 456,
        }
        self.assertEqual(res, resExpected)

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.dataTakingConditions = "dataTC"
        pr.processingPass = "procePass"
        pr.dqFlag = "OK,AA, BB"
        pr.bkFileType = ["type1"]
        pr.startRun = "123"
        pr.endRun = "456"
        res = pr._getBKKQuery()
        resExpected = {
            "DataTakingConditions": "dataTC",
            "ProcessingPass": "procePass",
            "FileType": "type1",
            "EventType": "",
            "ConfigName": "test",
            "ConfigVersion": "certification",
            "DataQualityFlag": "OK;;;AA;;;BB",
            "StartRun": 123,
            "EndRun": 456,
        }
        self.assertEqual(res, resExpected)

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.dataTakingConditions = "dataTC"
        pr.processingPass = "procePass"
        pr.dqFlag = "OK,AA, BB"
        pr.bkFileType = "type1"
        pr.startRun = "123"
        pr.endRun = "456"
        res = pr._getBKKQuery()
        resExpected = {
            "DataTakingConditions": "dataTC",
            "ProcessingPass": "procePass",
            "FileType": "type1",
            "EventType": "",
            "ConfigName": "test",
            "ConfigVersion": "certification",
            "DataQualityFlag": "OK;;;AA;;;BB",
            "StartRun": 123,
            "EndRun": 456,
        }
        self.assertEqual(res, resExpected)

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.dataTakingConditions = "dataTC"
        pr.processingPass = "procePass"
        pr.dqFlag = "OK,AA, BB"
        pr.bkFileType = "type1, type2"
        pr.startRun = "123"
        pr.endRun = "456"
        res = pr._getBKKQuery()
        resExpected = {
            "DataTakingConditions": "dataTC",
            "ProcessingPass": "procePass",
            "FileType": "type1;;;type2",
            "EventType": "",
            "ConfigName": "test",
            "ConfigVersion": "certification",
            "DataQualityFlag": "OK;;;AA;;;BB",
            "StartRun": 123,
            "EndRun": 456,
        }
        self.assertEqual(res, resExpected)

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.dqFlag = "OK,AA, BB"
        res = pr._getBKKQuery("fromPreviousProd", ["type"], 123)
        resExpected = {"ProductionID": 123, "FileType": "type", "DataQualityFlag": "OK;;;AA;;;BB"}
        self.assertEqual(res, resExpected)

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.dqFlag = "OK,AA, BB"
        res = pr._getBKKQuery("fromPreviousProd", ["type"], 123)
        resExpected = {"ProductionID": 123, "FileType": "type", "DataQualityFlag": "OK;;;AA;;;BB"}
        self.assertEqual(res, resExpected)

        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.dqFlag = "OK,AA, BB"
        res = pr._getBKKQuery("fromPreviousProd", ["type1", "type2"], 123)
        resExpected = {"ProductionID": 123, "FileType": "type1;;;type2", "DataQualityFlag": "OK;;;AA;;;BB"}
        self.assertEqual(res, resExpected)

    def test__splitIntoProductionSteps(self):
        r_p = _splitIntoProductionSteps(stepStripp)
        r_exp = [
            {
                "ApplicationName": "DaVinci",
                "Usable": "Yes",
                "StepId": 123,
                "ApplicationVersion": "v28r3p1",
                "ExtraPackages": "AppConfig.v3r104",
                "StepName": "Stripping14-Merging",
                "SystemConfig": "",
                "ProcessingPass": "Merging",
                "Visible": "N",
                "DDDB": "head-20110302",
                "mcTCK": "",
                "OptionFiles": "$APPCONFIGOPTS/Merging/DV-Stripping14-Merging.py",
                "CONDDB": "head-20110407",
                "visibilityFlag": [],
                "fileTypesIn": ["SDST"],
                "fileTypesOut": ["BHADRON.DST", "CALIBRATION.DST", "CHARM.MDST", "CHARMCOMPLETEEVENT.DST"],
            }
        ]

        self.assertEqual(r_p, r_exp)

        r_p = _splitIntoProductionSteps(mergeStep)
        r_exp = [
            {
                "ApplicationName": "DaVinci",
                "Usable": "Yes",
                "StepId": 456,
                "ApplicationVersion": "v28r3p1",
                "ExtraPackages": "AppConfig.v3r104",
                "StepName": "Stripping14-Merging",
                "SystemConfig": "",
                "prodStepID": "456['BHADRON.DST']",
                "mcTCK": "",
                "ProcessingPass": "Merging",
                "Visible": "N",
                "DDDB": "head-20110302",
                "OptionFiles": "$APPCONFIGOPTS/Merging/DV-Stripping14-Merging.py",
                "CONDDB": "head-20110407",
                "visibilityFlag": [],
                "fileTypesIn": ["BHADRON.DST"],
                "fileTypesOut": ["BHADRON.DST"],
            },
            {
                "ApplicationName": "DaVinci",
                "Usable": "Yes",
                "StepId": 456,
                "ApplicationVersion": "v28r3p1",
                "ExtraPackages": "AppConfig.v3r104",
                "StepName": "Stripping14-Merging",
                "SystemConfig": "",
                "prodStepID": "456['CALIBRATION.DST']",
                "mcTCK": "",
                "ProcessingPass": "Merging",
                "Visible": "N",
                "DDDB": "head-20110302",
                "OptionFiles": "$APPCONFIGOPTS/Merging/DV-Stripping14-Merging.py",
                "CONDDB": "head-20110407",
                "visibilityFlag": [],
                "fileTypesIn": ["CALIBRATION.DST"],
                "fileTypesOut": ["CALIBRATION.DST"],
            },
            {
                "ApplicationName": "DaVinci",
                "Usable": "Yes",
                "StepId": 456,
                "ApplicationVersion": "v28r3p1",
                "ExtraPackages": "AppConfig.v3r104",
                "StepName": "Stripping14-Merging",
                "SystemConfig": "",
                "prodStepID": "456['PID.MDST']",
                "mcTCK": "",
                "ProcessingPass": "Merging",
                "Visible": "N",
                "DDDB": "head-20110302",
                "OptionFiles": "$APPCONFIGOPTS/Merging/DV-Stripping14-Merging.py",
                "CONDDB": "head-20110407",
                "visibilityFlag": [],
                "fileTypesIn": ["PID.MDST"],
                "fileTypesOut": ["PID.MDST"],
            },
        ]

        self.assertEqual(r_p, r_exp)


class ProductionRequestFailure(ClientTestCase):
    def test_resolveStepsFailure(self):
        pr = ProductionRequest(self.bkClientMock, self.diracProdIn)
        pr.stepsList = ["123"]
        self.bkClientMock.getAvailableSteps.return_value = {"OK": False, "Message": "error"}
        self.assertRaises(ValueError, pr.resolveSteps)

        self.bkClientMock.getStepInputFiles.return_value = {"OK": False, "Message": "error"}
        self.assertRaises(ValueError, pr.resolveSteps)


class ProductionRequestFullChain(ClientTestCase):
    def test_full(self):
        pr = ProductionRequest(self.bkClientFake, self.diracProdIn)
        pr.logger.setLevel("VERBOSE")

        stepsList = ["123"]
        stepsList.append("456")
        stepsList.append("")
        stepsList.append("")
        stepsList.append("")
        pr.stepsList = stepsList
        pr.compressionLvl = ["LOW"] * 4
        pr.outputVisFlag = [{"1": "N"}, {"2": "N"}]
        pr.resolveSteps()

        pr.appendName = "1"
        pr.configName = "MC"
        pr.configVersion = "MC11a"

        pr.events = ["100", "-1"]

        pr.eventsToProduce = 10000

        pr.eventType = "11124001"
        pr.parentRequestID = "34"
        pr.requestID = "35"

        pr.prodGroup = "Sim05/Trig0x40760037Flagged/Reco12a/Stripping17Flagged"
        pr.dataTakingConditions = "Beam3500GeV-2011-MagDown-Nu2-EmNoCuts"

        pr.prodsTypeList = ["Stripping", "Merge"]
        pr.outputSEs = ["Tier1_MC-DST", "Tier1_MC-DST"]
        pr.specialOutputSEs = [{}, {}]
        pr.stepsInProds = [range(1, len(pr.stepsList)), [len(pr.stepsList)]]
        pr.removeInputsFlags = [False, True]
        pr.priorities = [1, 6]
        pr.cpus = [1000, 100]
        pr.outputModes = ["Local", "Any"]
        pr.outputFileMasks = ["FOO", ""]
        pr.outputFileSteps = ["2", ""]
        pr.targets = ["Tier2", ""]
        pr.groupSizes = [1, 5]
        pr.plugins = ["", "BySize"]
        pr.inputDataPolicies = ["", "protocol"]
        pr.inputs = [[], []]

        res = pr.buildAndLaunchRequest()

        self.assertEqual(res, {"OK": True, "Value": [321, 321, 321]})


# ###########################################################################
# Test Suite run
# ###########################################################################


if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(ClientTestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ProductionRequestSuccess))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ProductionRequestFailure))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ProductionRequestFullChain))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)

# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""NagiosConnector is a utility to publish checks on job results to the IT-
based SAM/Nagios framework.

The publishing is done by messaging via an ActiveMQ-Broker specified in
the configuration (It expects values for MsgBroker, MsgPort,  MsgQueue
and NagiosName). For the message to arrive, it is essential to publish
to the right queue, which is also specified in the configuration. The
message is built using a dictionary passed to one of the methods, which
should contain the keys 'SAMResults' 'SAMDetails' 'GridRequiredCEs'.
"""
import datetime
import stomp

from DIRAC import gLogger, gConfig
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations


class NagiosConnector:
    """Bundles functions in the stomp library for sending SAMJob-Results to SAM-
    Nagios."""

    def __init__(self):
        self.config = {}
        self.opsHelper = Operations()
        self.conn = None
        self.message = ""

    def readConfig(self):
        """get Message Broker and Queue from Configuration file."""

        self.config = {}
        for item in ["MsgBroker", "MsgQueue", "MsgPort"]:
            path = self.opsHelper.getPath(f"/NagiosConnector/{item}")
            self.config[item] = gConfig.getValue(path)
            if not self.config[item]:
                gLogger.verbose(f"Required Configuration Value is empty: {item}")
            else:
                gLogger.verbose(f"Read Config Values for {item}: {self.config[item]}")

        try:
            self.config["MsgPort"] = int(self.config["MsgPort"])
        except (TypeError, ValueError):
            self.config["MsgPort"] = 6163

    def useDebugMessage(self):
        """Load a sample message for debugging."""
        self.message = """hostName: ce.hpc.iit.bme.hu
metricStatus: OK
timestamp: 2013-11-09T17:59:19Z
nagiosName: org.lhcb.DiracTest-lhcb
summaryData: publishing by Dirac successful
serviceURI:   ce.hpc.iit.bme.hu
serviceFlavour: CE
siteName: myTestSite
metricName: org.lhcb.DiracTest
gatheredAt: sam-developers-machine
role: site
voName: lhcb
serviceType: org.lhcb.CE
detailsData: further details here:
EOT"""

    def assembleMessage(
        self,
        serviceURI="no serviceURI given",
        status="no status given",
        details="no details given",
        nagiosName="no nagiosName given",
    ):
        """Brings message information to the generic format required by Nagios."""

        statuscodes = {0: "OK", 1: "CRITICAL", 2: "WARNING", 3: "UNKNOWN"}
        if status in ["CRITICAL", "OK", "WARNING", "UNKNOWN"]:
            pass
        elif status in range(4):
            status = statuscodes[status]

            currentTime = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")
            message = ""
            message += f"serviceURI: {serviceURI}\n"
            message += f"metricStatus: {status}\n"
            message += f"timestamp: {currentTime}\n"
            message += f"nagiosName: {nagiosName}\n"
            message += f"summaryData: {details}\n"
            # all other fields are not used in SAM-Nagios
            # they are left here for compatibility / legacy
            # message += "gatheredAt: %s\n" % msginfos['ce']
            # message += "metricName: org.lhcb.WN-%s-lhcb\n" % msginfos[ 'JobName' ]
            # message += "serviceFlavour: %s\n" % msginfos['serviceFlavour']
            # message += "role: site\n"
            # message += "voName: lhcb\n"
            # message += "serviceType: org.lhcb.WN\n"
            # message += "detailsData: %s\n" % msginfos[ 'testDetails' ]
            message += "EOT\n"
            self.message = message

    def initializeConnection(
        self, _use_ssl=False, _ssl_key_file=None, _ssl_cert_file=None, _ssl_ca_certs=None, _ssl_cert_validator=None
    ):
        """Connect the conn object with the Broker read from configuration.

        Refer to the stomppy documentation for full authentication args details.

        :param _use_ssl:  connect using SSL to the socket.
          This wraps the socket in a SSL connection.
          The constructor will raise an exception if you ask for SSL,
          but it can't find the SSL module.
        :param _ssl_cert_file:  the path to a X509 certificate
        :param _ssl_key_file: the path to a X509 key file
        :param _ssl_ca_certs:  the path to the a file containing CA certificates to validate the server against.
        :param _ssl_cert_validator:  function which performs extra validation on the client certificate
        """
        try:
            self.conn = stomp.Connection(
                [(self.config["MsgBroker"], self.config["MsgPort"])],
            )
            self.conn.set_ssl(
                use_ssl=_use_ssl,
                ssl_key_file=_ssl_key_file,
                ssl_cert_file=_ssl_cert_file,
                ssl_ca_certs=_ssl_ca_certs,
                ssl_cert_validator=_ssl_cert_validator,
            )
            # There may be a need to receive messages.
            # In this case there should be a class with an on_message method
            # conn.set_listener('',MyListener()) python-messaging provides a useful class.
            self.conn.connect()
        except stomp.exception.ConnectFailedException as e:
            gLogger.error(f"Error establishing connection: {e}")

    def sendMessage(self):
        """Use the conn object to send a message to the broker.

        If the format and the configurations are correct, the message
        content will appear in SAM/Nagios.
        """

        if not self.message:
            gLogger.error("The message string is empty!")
        self.conn.send(destination=self.config["MsgQueue"], body=self.message)
        gLogger.verbose(f"Message sent to {self.config['MsgBroker']} on {self.config['MsgQueue']}")
        gLogger.verbose(f"Message content {self.message}")

    def endConnection(self):
        """Call disconnect() on the conn object."""
        self.conn.disconnect()
        gLogger.verbose("Connection successfully terminated")



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Test_Resources_Storage_LHCbOnlineStorage."""
from unittest import mock
import unittest

import LHCbDIRAC.Resources.Storage.LHCbOnlineStorage as moduleTested

################################################################################


class LHCbOnlineStorage_TestCase(unittest.TestCase):
    def setUp(self):
        """Setup."""

        # Mock external libraries / modules not interesting for the unit test
        mock_xmlrpclib = mock.Mock()
        mock_xmlrpclib.Server.return_value("")
        self.mock_xmlrpclib = mock_xmlrpclib

        # Add mocks to moduleTested
        moduleTested.xmlrpclib = self.mock_xmlrpclib

        self.moduleTested = moduleTested
        self.testClass = self.moduleTested.LHCbOnlineStorage

    def tearDown(self):
        """TearDown."""
        del self.testClass
        del self.moduleTested
        del self.mock_xmlrpclib


################################################################################
# Tests


class LHCbOnlineStorage_Success(LHCbOnlineStorage_TestCase):
    def test_instantiate(self):
        """tests that we can instantiate one object of the tested class."""

        resource = self.testClass(
            "storageName",
            {
                "Protocol": "protocol",
                "Path": "path",
                "Host": "host",
                "Port": "port",
                "SpaceToken": "spaceToken",
                "WSUrl": "wspath",
            },
        )

        self.assertEqual("LHCbOnlineStorage", resource.__class__.__name__)

    def test_init(self):
        """tests that the init method does what it should do."""

        resource = self.testClass(
            "storageName",
            {
                "Protocol": "protocol",
                "Path": "path",
                "Host": "host",
                "Port": "port",
                "SpaceToken": "spaceToken",
                "WSUrl": "wspath",
            },
        )

        self.assertEqual("storageName", resource.name)
        self.assertEqual("protocol", resource.protocolParameters["Protocol"])
        self.assertEqual("path", resource.protocolParameters["Path"])
        self.assertEqual("host", resource.protocolParameters["Host"])
        self.assertEqual("port", resource.protocolParameters["Port"])
        self.assertEqual("spaceToken", resource.protocolParameters["SpaceToken"])
        self.assertEqual("wspath", resource.protocolParameters["WSUrl"])

    def test_getParameters(self):
        """tests the output of getParameters method."""

        resource = self.testClass(
            "storageName",
            {
                "Protocol": "protocol",
                "Path": "path",
                "Host": "host",
                "Port": "port",
                "SpaceToken": "spaceToken",
                "WSUrl": "wspath",
            },
        )

        res = resource.getParameters()

        self.assertEqual("storageName", res["StorageName"])
        self.assertEqual("protocol", res["Protocol"])
        self.assertEqual("path", res["Path"])
        self.assertEqual("host", res["Host"])
        self.assertEqual("port", res["Port"])
        self.assertEqual("spaceToken", res["SpaceToken"])
        self.assertEqual("wspath", res["WSUrl"])

    def test_getFileSize(self):
        """tests the output of getFileSize."""

        resource = self.testClass(
            "storageName",
            {
                "Protocol": "protocol",
                "Path": "path",
                "Host": "host",
                "Port": "port",
                "SpaceToken": "spaceToken",
                "WSUrl": "wspath",
            },
        )

        res = resource.getFileSize({})
        self.assertEqual(False, res["OK"])

        res = resource.getFileSize([])
        self.assertEqual(False, res["OK"])

        res = resource.getFileSize(["A", "B"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"A": 0, "B": 0}, res["Value"]["Successful"])
        self.assertEqual({}, res["Value"]["Failed"])

        res = resource.getFileSize({"A": 1, "B": {}})
        self.assertEqual(True, res["OK"])
        self.assertEqual({"A": 0, "B": 0}, res["Value"]["Successful"])
        self.assertEqual({}, res["Value"]["Failed"])

    def test_removeFile(self):
        """tests the output of removeFile."""

        resource = self.testClass(
            "storageName",
            {
                "Protocol": "protocol",
                "Path": "path",
                "Host": "host",
                "Port": "port",
                "SpaceToken": "spaceToken",
                "WSUrl": "wspath",
            },
        )

        # no need for mock here, we don't reach it

        res = resource.removeFile({})
        self.assertEqual(False, res["OK"])

        res = resource.removeFile([])
        self.assertEqual(False, res["OK"])

        resource.server.endMigratingFileBulk.return_value = (1, {})

        res = resource.removeFile(["A", "B"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"A": True, "B": True}, res["Value"]["Successful"])
        self.assertEqual({}, res["Value"]["Failed"])

        resource.server.endMigratingFileBulk.return_value = (1, {"B": "Did not work"})

        res = resource.removeFile(["A", "B"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"A": True}, res["Value"]["Successful"])
        self.assertEqual(["B"], list(res["Value"]["Failed"]))

        resource.server.endMigratingFileBulk.side_effect = Exception("Boom!")
        res = resource.removeFile(["A", "B"])
        # FIXME: This should return S_ERROR !!
        self.assertEqual(True, res["OK"])
        self.assertEqual({}, res["Value"]["Successful"])
        self.assertEqual(["A", "B"], list(res["Value"]["Failed"]))

    def test_retransferOnlineFile(self):
        """tests output of retransferOnlineFile."""

        resource = self.testClass(
            "storageName",
            {
                "Protocol": "protocol",
                "Path": "path",
                "Host": "host",
                "Port": "port",
                "SpaceToken": "spaceToken",
                "WSUrl": "wspath",
            },
        )
        resource.server.errorMigratingFile.return_value = (1, 0)

        res = resource.retransferOnlineFile({})
        self.assertEqual(False, res["OK"])

        res = resource.retransferOnlineFile([])
        self.assertEqual(False, res["OK"])

        res = resource.retransferOnlineFile(["A", "B"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"A": True, "B": True}, res["Value"]["Successful"])
        self.assertEqual({}, res["Value"]["Failed"])

        # side_effect does not work very well, cooked a workaround
        _myValues = [(0, 1), (1, 0)]

        def _side_effect(_pfn):
            return _myValues.pop()

        resource.server.errorMigratingFile.side_effect = _side_effect

        res = resource.retransferOnlineFile(["A", "B"])
        self.assertEqual(True, res["OK"])
        self.assertEqual({"A": True}, res["Value"]["Successful"])
        self.assertEqual(["B"], list(res["Value"]["Failed"]))

        resource.server.errorMigratingFile.side_effect = Exception("Boom!")
        res = resource.retransferOnlineFile(["A", "B"])
        # FIXME: This should return S_ERROR !!
        self.assertEqual(True, res["OK"])
        self.assertEqual({}, res["Value"]["Successful"])
        self.assertEqual(["A", "B"], list(res["Value"]["Failed"]))



#!/usr/bin/env python
###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""remove output of production."""
import sys

from DIRAC.Core.Base.Script import Script


@Script()
def main():
    from DIRAC.Core.Base.Script import parseCommandLine

    parseCommandLine()

    from DIRAC import exit as DIRACExit
    from DIRAC import gLogger
    from LHCbDIRAC.TransformationSystem.Agent.TransformationCleaningAgent import TransformationCleaningAgent
    from LHCbDIRAC.TransformationSystem.Client.TransformationClient import TransformationClient

    transIDs = []
    if len(sys.argv) < 2:
        print("Usage: dirac-production-remove-output transID [transID] [transID]")
        DIRACExit(1)
    else:
        try:
            transIDs = [int(arg) for arg in sys.argv[1:]]
        except BaseException:
            print("Invalid list of productions")
            DIRACExit(1)

    agent = TransformationCleaningAgent(
        "Transformation/TransformationCleaningAgent",
        "Transformation/TransformationCleaningAgent",
        "dirac-production-remove-output",
    )
    agent.initialize()

    client = TransformationClient()
    for transID in transIDs:
        res = client.getTransformationParameters(transID, ["Status"])
        if not res["OK"]:
            gLogger.error("Failed to determine transformation status")
            gLogger.error(res["Message"])
            continue
        status = res["Value"]
        if status not in ["RemovingFiles", "RemovingOutput", "ValidatingInput", "Active"]:
            gLogger.error(f"The transformation is in {status} status and the outputs cannot be removed")
            continue
        agent.removeTransformationOutput(transID)


if __name__ == "__main__":
    main()



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Simple merging module for MDF files."""
import shlex

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.Core.Utilities.Subprocess import systemCall
from DIRAC.Resources.Catalog.PoolXMLCatalog import PoolXMLCatalog

import LHCbDIRAC
from LHCbDIRAC.Workflow.Modules.ModuleBase import ModuleBase


class MergeMDF(ModuleBase):
    """To be used in normal workflows."""

    #############################################################################
    def __init__(self, bkClient=None, dm=None):
        """Module initialization."""
        self.log = gLogger.getSubLogger("MergeMDF")
        super().__init__(self.log, bkClientIn=bkClient, dm=dm)

        self.outputLFN = ""
        # List all input parameters here
        self.stepInputData = []
        self.poolXMLCatName = "pool_xml_catalog.xml"
        self.applicationName = "cat"

    #############################################################################
    def _resolveInputVariables(self):
        """By convention the module parameters are resolved here."""

        super()._resolveInputVariables()
        super()._resolveInputStep()

    #############################################################################

    def execute(
        self,
        production_id=None,
        prod_job_id=None,
        wms_job_id=None,
        workflowStatus=None,
        stepStatus=None,
        wf_commons=None,
        step_commons=None,
        step_number=None,
        step_id=None,
    ):
        """Main execution function."""

        try:
            super().execute(
                production_id,
                prod_job_id,
                wms_job_id,
                workflowStatus,
                stepStatus,
                wf_commons,
                step_commons,
                step_number,
                step_id,
            )

            poolCat = PoolXMLCatalog(self.poolXMLCatName)

            self._resolveInputVariables()

            stepOutputs, stepOutputTypes, _histogram = self._determineOutputs()

            logLines = ["#" * len(LHCbDIRAC.version), LHCbDIRAC.version, "#" * len(LHCbDIRAC.version)]

            localInputs = [str(list(poolCat.getPfnsByLfn(x)["Replicas"].values())[0]) for x in self.stepInputData]
            inputs = " ".join(localInputs)
            cmd = f"cat {inputs} > {self.outputFilePrefix + '.' + stepOutputTypes[0]}"
            logLines.append("\nExecuting merge operation...")
            self.log.info(f'Executing "{cmd}"')
            result = systemCall(timeout=600, cmdSeq=shlex.split(cmd))
            if not result["OK"]:
                self.log.error(result)
                logLines.append(f"Merge operation failed with result:\n{result}")
                return S_ERROR("Problem Executing Application")

            status = result["Value"][0]
            stdout = result["Value"][1]
            stderr = result["Value"][2]
            self.log.info(stdout)
            if stderr:
                self.log.error(stderr)

            if status:
                msg = f'Non-zero status {status} while executing "{cmd}"'
                self.log.info(msg)
                logLines.append(msg)
                return S_ERROR("Problem Executing Application")

            self.log.info(f"Going to manage {self.applicationName} output")
            self._manageAppOutput(stepOutputs)

            # Still have to set the application status e.g. user job case.
            self.setApplicationStatus(f"{self.applicationName} {self.applicationVersion} Successful")

            # Write to log file
            msg = "Produced merged MDF file"
            self.log.info(msg)
            logLines.append(msg)
            logLines = [str(i) for i in logLines]
            logLines.append("#EOF")
            with open(self.applicationLog, "w") as fopen:
                fopen.write("\n".join(logLines) + "\n")

            return S_OK(f"{self.applicationName} {self.applicationVersion} Successful")

        except Exception as e:  # pylint:disable=broad-except
            self.log.exception("Failure in MergeMDF execute module", lException=e)
            return S_ERROR(str(e))

        finally:
            super().finalize()


# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#



###############################################################################
# (c) Copyright 2022 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import logging

import pytest

from DIRAC.Core.Base.AgentModule import AgentModule

from LHCbDIRAC.ProductionManagementSystem.Agent import HistogramMergingAgent as hma_module
from LHCbDIRAC.ProductionManagementSystem.Agent.HistogramMergingAgent import HistogramMergingAgent, AGENT_NAME

NUM_FILES = 10


def example_transformations():
    return [
        {
            "TransformationID": 1234,
            "TransformationFamily": "567",
            "Type": "MCSimulation",
            "Body": """<Workflow>
            <Parameter name="configName" type="string" linked_module="" linked_parameter="" in="True" out="False" description="ConfigName"><value><![CDATA[MC]]></value></Parameter>
            <Parameter name="configVersion" type="string" linked_module="" linked_parameter="" in="True" out="False" description="ConfigVersion"><value><![CDATA[2018]]></value></Parameter>
            <Parameter name="eventType" type="string" linked_module="" linked_parameter="" in="True" out="False" description="Event Type of the production"><value><![CDATA[13264021]]></value></Parameter>
            <StepInstance>
            <name>Gauss_1</name>
            <Parameter name="applicationName" type="string" linked_module="" linked_parameter="" in="True" out="False" description="ApplicationName"><value><![CDATA[Gauss]]></value></Parameter>
            <Parameter name="optionsFormat" type="string" linked_module="" linked_parameter="" in="True" out="False" description="ProdConf configuration"><value><![CDATA[]]></value></Parameter>
            </StepInstance>
            </Workflow>
            """,  # noqa # pylint: disable=line-too-long
        },
        {
            "TransformationID": 1235,
            "TransformationFamily": "567",
            "Type": "MCReconstruction",
            "Body": """<Workflow>
            <Parameter name="configName" type="string" linked_module="" linked_parameter="" in="True" out="False" description="ConfigName"><value><![CDATA[MC]]></value></Parameter>
            <Parameter name="configVersion" type="string" linked_module="" linked_parameter="" in="True" out="False" description="ConfigVersion"><value><![CDATA[2018]]></value></Parameter>
            <Parameter name="eventType" type="string" linked_module="" linked_parameter="" in="True" out="False" description="Event Type of the production"><value><![CDATA[13264021]]></value></Parameter>
            <StepInstance>
            <name>Boole_1</name>
            <Parameter name="applicationName" type="string" linked_module="" linked_parameter="" in="True" out="False" description="ApplicationName"><value><![CDATA[Boole]]></value></Parameter>
            <Parameter name="optionsFormat" type="string" linked_module="" linked_parameter="" in="True" out="False" description="ProdConf configuration"><value><![CDATA[]]></value></Parameter>
            </StepInstance>
            <StepInstance>
            <name>Moore_2</name>
            <Parameter name="applicationName" type="string" linked_module="" linked_parameter="" in="True" out="False" description="ApplicationName"><value><![CDATA[Moore]]></value></Parameter>
            <Parameter name="optionsFormat" type="string" linked_module="" linked_parameter="" in="True" out="False" description="ProdConf configuration"><value><![CDATA[l0app]]></value></Parameter>
            </StepInstance>
            <StepInstance>
            <name>Moore_3</name>
            <Parameter name="applicationName" type="string" linked_module="" linked_parameter="" in="True" out="False" description="ApplicationName"><value><![CDATA[Moore]]></value></Parameter>
            <Parameter name="optionsFormat" type="string" linked_module="" linked_parameter="" in="True" out="False" description="ProdConf configuration"><value><![CDATA[]]></value></Parameter>
            </StepInstance>
            <StepInstance>
            <name>Moore_4</name>
            <Parameter name="applicationName" type="string" linked_module="" linked_parameter="" in="True" out="False" description="ApplicationName"><value><![CDATA[Moore]]></value></Parameter>
            <Parameter name="optionsFormat" type="string" linked_module="" linked_parameter="" in="True" out="False" description="ProdConf configuration"><value><![CDATA[]]></value></Parameter>
            </StepInstance>
            <StepInstance>
            <name>Brunel_5</name>
            <Parameter name="applicationName" type="string" linked_module="" linked_parameter="" in="True" out="False" description="ApplicationName"><value><![CDATA[Brunel]]></value></Parameter>
            <Parameter name="optionsFormat" type="string" linked_module="" linked_parameter="" in="True" out="False" description="ProdConf configuration"><value><![CDATA[]]></value></Parameter>
            </StepInstance>
            <StepInstance>
            <name>DaVinci_6</name>
            <Parameter name="applicationName" type="string" linked_module="" linked_parameter="" in="True" out="False" description="ApplicationName"><value><![CDATA[DaVinci]]></value></Parameter>
            <Parameter name="optionsFormat" type="string" linked_module="" linked_parameter="" in="True" out="False" description="ProdConf configuration"><value><![CDATA[Tesla]]></value></Parameter>
            </StepInstance>
            <StepInstance>
            <name>DaVinci_7</name>
            <Parameter name="applicationName" type="string" linked_module="" linked_parameter="" in="True" out="False" description="ApplicationName"><value><![CDATA[DaVinci]]></value></Parameter>
            <Parameter name="optionsFormat" type="string" linked_module="" linked_parameter="" in="True" out="False" description="ProdConf configuration"><value><![CDATA[]]></value></Parameter>
            </StepInstance>
            </Workflow>
            """,  # noqa # pylint: disable=line-too-long
        },
    ]


STEPS = {
    "MCSimulation": ["Gauss"],
    "MCReconstruction": [
        "Boole",
        "Moore",  # L0
        "Moore",  # Hlt1
        "Moore",  # Hlt2
        "Brunel",
        "DaVinci",  # Turbo
        "DaVinci",  # Stripping
    ],
}

EXPECTED_YAML_FILES = {
    "MCSimulation": [
        "1-GAUSSHIST.yml",
    ],
    "MCReconstruction": [
        "1-BOOLEHIST.yml",
        "2-L0APPHIST.yml",
        "3-MOOREHIST.yml",
        "4-MOOREHIST.yml",
        "5-BRUNELHIST.yml",
        "6-TESLAHIST.yml",
        "7-DAVINCIHIST.yml",
    ],
}


def expected_modified_transformations():
    return [
        {
            "TransformationID": 1234,
            "TransformationFamily": "567",
            "Type": "MCSimulation",
            "Parameters": {
                "configName": "MC",
                "configVersion": "2018",
                "eventType": "13264021",
            },
            "Steps": {
                "Gauss_1": {
                    "applicationName": "Gauss",
                    "optionsFormat": None,
                }
            },
            "HistogramMerging": {
                "1-GAUSSHIST": {
                    "InputLFNs": [
                        f"/lhcb/MC/2018/GAUSSHIST/00001234/0000/Gauss_00001234_{job:08}_1.Hist.root"
                        for job in range(NUM_FILES)
                    ],
                    "InputPFNs": [
                        "root://eoslhcb.cern.ch//eos/lhcb/grid/prod/lhcb/MC/2018/GAUSSHIST/00001234/0000/"
                        f"Gauss_00001234_{job:08}_1.Hist.root"
                        for job in range(NUM_FILES)
                    ],
                    "OutputLFN": "/lhcb/MC/2018/GAUSSHISTMERGED/00001234/0000/Gauss_00001234_0_1.Hist.root",
                },
            },
        },
        {
            "TransformationID": 1235,
            "TransformationFamily": "567",
            "Type": "MCReconstruction",
            "Parameters": {
                "configName": "MC",
                "configVersion": "2018",
                "eventType": "13264021",
            },
            "Steps": {
                "Boole_1": {
                    "applicationName": "Boole",
                    "optionsFormat": None,
                },
                "Moore_2": {
                    "applicationName": "Moore",
                    "optionsFormat": "l0app",
                },
                "Moore_3": {
                    "applicationName": "Moore",
                    "optionsFormat": None,
                },
                "Moore_4": {
                    "applicationName": "Moore",
                    "optionsFormat": None,
                },
                "Brunel_5": {
                    "applicationName": "Brunel",
                    "optionsFormat": None,
                },
                "DaVinci_6": {
                    "applicationName": "DaVinci",
                    "optionsFormat": "Tesla",
                },
                "DaVinci_7": {
                    "applicationName": "DaVinci",
                    "optionsFormat": None,
                },
            },
            "HistogramMerging": {
                f"{step_index}-{app.upper()}HIST": {
                    "InputLFNs": [
                        f"/lhcb/MC/2018/{app.upper()}HIST/00001235/0000/{app}_00001235_{job:08}_{step_index}.Hist.root"
                        for job in range(NUM_FILES)
                    ],
                    "InputPFNs": [
                        f"root://eoslhcb.cern.ch//eos/lhcb/grid/prod/lhcb/MC/2018/{app.upper()}HIST/00001235/0000/"
                        f"{app}_00001235_{job:08}_{step_index}.Hist.root"
                        for job in range(NUM_FILES)
                    ],
                    "OutputLFN": f"/lhcb/MC/2018/{app.upper()}HISTMERGED/00001235/0000/"
                    f"{app}_00001235_0_{step_index}.Hist.root",
                }
                for step_index, app in enumerate(STEPS["MCReconstruction"], start=1)
            },
        },
    ]


def example_lfns():
    return [
        "/".join(
            [
                "/lhcb",
                transform["Parameters"]["configName"],
                transform["Parameters"]["configVersion"],
                f"{app.upper()}HIST",
                f"{transform['TransformationID']:08}",
                "0000",
                f"{app}_{transform['TransformationID']:08}_{job:08}_{step_index}.Hist.root",
            ]
        )
        for transform in expected_modified_transformations()
        for step_index, app in enumerate(STEPS[transform["Type"]], start=1)
        for job in range(NUM_FILES)
    ]


def example_requests():
    return {
        567: {
            "RequestID": "567",
            "RequestName": "",
            "RequestWG": "",
            "SimCondition": "",
            "SimCondDetail": "",
            "ProPath": "",
            "Extra": "",
            "RetentionRate": "",
            "FastSimulationType": "",
        }
    }


@pytest.fixture
def hma(monkeypatch):
    monkeypatch.setattr(AgentModule, "__init__", lambda *a, **kw: None)
    monkeypatch.setattr(AgentModule, "log", logging, raising=False)
    monkeypatch.setattr(AgentModule, "am_getOption", lambda _self, _name, default_value: default_value)
    # Values of module-level global variables that we want to change for testing
    global_values = {
        "gStandAlone": True,
        "gLocalWorkDir": True,
        "gDoRealMerging": False,
        "gDoRealUpdate": False,
        "gUseGitRepo": True,  # Can remain True while SimDQ repo is public, but we may run into issue of authentication later
    }
    for var, value in global_values.items():
        monkeypatch.setattr(hma_module, var, value)
    # Patch some of the query functions
    patched_functions = {
        "_latest_lbconda_env_version": "latest",  # original has @property dectorator
        "_query_recent_transformations": lambda _self: example_transformations(),
        "_query_requests": lambda _self, _request_ids: example_requests(),
        "_query_input_lfns": lambda _self, _transform_ids: example_lfns(),
    }
    for name, func in patched_functions.items():
        monkeypatch.setattr(HistogramMergingAgent, name, func)
    # Initialise the agent
    _hma = HistogramMergingAgent(AGENT_NAME, AGENT_NAME)
    _hma.initialize()
    _hma.beginExecution()
    yield _hma


def test_grouping(hma):
    """Test HistogramMergingAgent._group_lfns_by_prod_and_type"""
    expected = {
        transform["TransformationID"]: {
            file_key: merge_info["InputLFNs"] for file_key, merge_info in transform["HistogramMerging"].items()
        }
        for transform in expected_modified_transformations()
    }
    actual = hma._group_lfns_by_prod_and_type(example_lfns())  # pylint: disable=protected-access
    assert actual == expected


@pytest.mark.parametrize(
    "transform, expected_transform", zip(example_transformations(), expected_modified_transformations())
)
def test_xml_parsing(hma, transform, expected_transform):
    """Test HistogramMergingAgent._get_transformation_parameters"""
    workflow = hma._parse_workflow_xml(transform["Body"])  # pylint: disable=protected-access
    actual = hma._get_transformation_parameters(workflow)  # pylint: disable=protected-access
    expected = expected_transform["Parameters"]
    assert actual == expected


@pytest.mark.parametrize("transform", expected_modified_transformations())
def test_output_lfn(hma, transform):
    """Test HistogramMergingAgent._format_output_lfn"""
    for file_key, merge_info in transform["HistogramMerging"].items():
        actual = hma._format_output_lfn(transform, file_key)  # pylint: disable=protected-access
        expected = merge_info["OutputLFN"]
        assert actual == expected


def test_fetch_transformations(hma):
    """Test HistogramMergingAgent._fetch_transformations"""
    actual = hma._fetch_transformations()  # pylint: disable=protected-access
    expected = expected_modified_transformations()
    assert actual == expected


@pytest.mark.parametrize(
    "transform, production_request",
    zip(
        expected_modified_transformations(),
        [example_requests()[int(t["TransformationFamily"])] for t in expected_modified_transformations()],
    ),
)
def test_handle_transformation(hma, transform, production_request):
    """Test HistogramMergingAgent._handle_transformation"""
    new_yaml_files = [
        hma._handle_transformation(transform, production_request, file_key)  # pylint: disable=protected-access
        for file_key in transform["HistogramMerging"]
    ]
    merge_dir = hma.simdq_dir / transform["TransformationFamily"].rjust(8, "0") / transform["Parameters"]["eventType"]
    for actual, expected_fn in zip(new_yaml_files, EXPECTED_YAML_FILES[transform["Type"]]):
        expected = merge_dir / expected_fn
        assert actual == expected
        # TODO: check the contents of the file?


def test_execute(hma):
    """Test HistogramMergingAgent.execute"""
    ret = hma.execute()
    assert ret["OK"]
    for transform in expected_modified_transformations():
        merge_dir = (
            hma.simdq_dir / transform["TransformationFamily"].rjust(8, "0") / transform["Parameters"]["eventType"]
        )
        for yaml_fn in EXPECTED_YAML_FILES[transform["Type"]]:
            yaml_file = merge_dir / yaml_fn
            assert yaml_file.exists()
    # TODO: check the output



###############################################################################
# (c) Copyright 2019 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Utilities to parse the XML Generator Logs."""
import ast
import json
import xmltodict
import re


def counterJson(listCounters):
    """returns a dictionary containing counters

    :param list listCounters: list containing all the counter nodes
    """
    dictCounters = dict()
    if isinstance(listCounters, dict):
        listCounters = [listCounters]
    for counter in listCounters:
        dictCounters[counter["@name"]] = int(counter["value"])
    return dictCounters


def efficiencyJson(listEfficiencies):
    """returns a dictionary containing efficiencies

    :param list listEfficiencies: list containing all the efficiency nodes
    """
    dictEfficiencies = dict()
    if isinstance(listEfficiencies, dict):
        listEfficiencies = [listEfficiencies]
    for efficiency in listEfficiencies:
        dictEfficiencies[efficiency["@name"]] = {
            "after": int(efficiency["after"]),
            "before": int(efficiency["before"]),
            "error": float(efficiency["error"]),
            "value": float(efficiency["value"]),
        }
    return dictEfficiencies


def fractionJson(listFractions):
    """returns a dictionary containing fractions

    :param list listFractions: list containing all the fraction nodes
    """
    dictFractions = dict()
    if isinstance(listFractions, dict):
        listFractions = [listFractions]
    for fraction in listFractions:
        dictFractions[fraction["@name"]] = {
            "number": int(fraction["number"]),
            "error": float(fraction["error"]),
            "value": float(fraction["value"]),
        }
    return dictFractions


def crossSectionJson(listCrossSections):
    """returns a dictionary containing cross sections

    :param list listCrossSections: list containing all the cross section nodes
    """
    dictCrossSections = dict()
    if isinstance(listCrossSections, dict):
        listCrossSections = [listCrossSections]
    for crossSection in listCrossSections:
        dictCrossSections[crossSection["description"][1:-1]] = {
            "ID": int(crossSection["@id"]),
            "generated": int(crossSection["generated"]),
            "value": float(crossSection["value"]),
        }
    return dictCrossSections


def methodGeneratorJson(listMethods, listGenerators, numberEventTypes):
    """returns a dictionary containing the generator for each method

    :param list listMethods: list containing all the methods
    :param list listGenerators: list containing all the generators
    :param int numberEventTypes: number of event types in the generator log
    """
    dictMethods = dict()
    if numberEventTypes > 1:
        for i, method in enumerate(listMethods):
            dictMethods[method] = listGenerators[i]
    else:
        dictMethods[listMethods] = listGenerators
    return dictMethods


class GeneratorLog:

    correctiveRegExps = {
        r"[\x00-\x09\x0b\x0c\x0e-\x1f\x7f-\xff]+": "",  # correct for memory garbage from Fortran
        r"-?nan": "-1",  # all NaNs replaced with same invalid value
        r"\?+": "",  # account for replace chars when filtering ASCII
        r'>\s*?"\s*+': ">",  # eliminate padding in XS description tags
        r'(?<=<description>)\s*?"\s*': "",
        r'\s++\S{,2}"\s*(?=</description>)': "",
        r'"\s(?=</description>)': "",
    }

    def __init__(self):
        pass

    def generatorLogJson(self, fileName):
        """converts the xml Generator Log into json format"""
        dictElements = dict()
        dictGenerator = dict()

        with open("GeneratorLog.xml", "rb") as fp:
            fileLines = fp.read().decode("ascii", errors="replace").split("\n")

        fileLines = fileLines[fileLines.index("<generatorCounters>") : fileLines.index("</generatorCounters>") + 1]
        xmlText = "".join(fileLines)
        numberEventTypes = xmlText.count("<eventType>")
        if numberEventTypes > 1:
            # Taking the first set of nodes
            xmlText = (
                xmlText.split("<eventType>")[0]
                + "<eventType>"
                + xmlText.split("<eventType>")[1]
                + "<method>"
                + xmlText.split("<eventType>")[-1].split("<method>", 1)[-1]
            )  # noqa

        for rexp, rrepl in GeneratorLog.correctiveRegExps.items():
            xmlText = re.sub(rexp, rrepl, xmlText, re.I | re.M)

        dicto = xmltodict.parse(xmlText)
        jsonData = ast.literal_eval(json.dumps(dicto))

        listCounters = jsonData["generatorCounters"].get("counter", [])
        listEfficiencies = jsonData["generatorCounters"].get("efficiency", [])
        listFractions = jsonData["generatorCounters"].get("fraction", [])
        listCrossSections = jsonData["generatorCounters"].get("crosssection", [])
        listMethods = jsonData["generatorCounters"]["method"]
        listGenerators = jsonData["generatorCounters"]["generator"]

        dictElements["counter"] = counterJson(listCounters)
        dictElements["efficiency"] = efficiencyJson(listEfficiencies)
        dictElements["fraction"] = fractionJson(listFractions)
        dictElements["crossSection"] = crossSectionJson(listCrossSections)
        dictElements["method"] = methodGeneratorJson(listMethods, listGenerators, numberEventTypes)

        dictGenerator["generatorCounters"] = dictElements

        with open(fileName, "w", encoding="utf-8") as fp:
            fp.write(str(json.dumps(dictGenerator, indent=2)))

        return dictGenerator
